/*
 * Originally adapted by Gary Thomas.  Much additional work by
 * Cort Dougan <cort@fsmlabs.com>.  On top of that still more work by
 * Dan Malek <dmalek@jlc.net>.
 *
 * Currently maintained by: Tom Rini <trini@kernel.crashing.org>
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/elf.h>
#include <linux/string.h>
#include <asm/bootinfo.h>
#include <asm/mmu.h>
#include <asm/page.h>
#include <asm/processor.h>
#include <asm/residual.h>
#if defined(CONFIG_4xx)
#include <asm/ibm4xx.h>
#elif defined(CONFIG_8xx)
#include <asm/mpc8xx.h>
#elif defined(CONFIG_8260)
#include <asm/mpc8260.h>
#endif
#include "eps.h"
#include "nonstdio.h"

#ifdef CONFIG_SILKWORM
#include <asm/bootinfo.h>
#include "brocade/sin.h"
#endif

/* The linker tells us where the image is. */
extern char __image_begin, __image_end;
extern char __ramdisk_begin, __ramdisk_end;
extern char _end[];

/* Because of the limited amount of memory on embedded, it presents
 * loading problems.  The biggest is that we load this boot program
 * into a relatively low memory address, and the Linux kernel Bss often
 * extends into this space when it get loaded.  When the kernel starts
 * and zeros the BSS space, it also writes over the information we
 * save here and pass to the kernel (usually board info).
 * On these boards, we grab some known memory holes to hold this information.
 */
char cmd_buf[256];
char *cmd_line = cmd_buf;
char *avail_ram;
char *end_avail;
char *zimage_start;

/* This is for 4xx treeboot.  It provides a place for the bootrom
 * give us a pointer to a rom environment command line.
 */
char *bootrom_cmdline = "";
char treeboot_bootrom_cmdline[512];

/* This is the default cmdline that will be given to the user at boot time..
 * If none was specified at compile time, we'll give it one that should work.
 * -- Tom */
#ifdef CONFIG_CMDLINE_BOOL
char compiled_string[] = CONFIG_CMDLINE;
#endif
char ramroot_string[] = "root=/dev/ram";
char netroot_string[] = "root=/dev/nfs rw ip=on";

/* Serial port to use. */
unsigned long com_port;

/* We need to make sure that this is before the images to ensure
 * that it's in a mapped location. - Tom */
bd_t hold_resid_buf __attribute__ ((__section__ (".data.boot")));
bd_t *hold_residual = &hold_resid_buf;

extern unsigned long serial_init(int chan, bd_t *bp);
extern void serial_close(unsigned long com_port);
extern unsigned long start;
extern void flush_instruction_cache(void);
extern void gunzip(void *, int, unsigned char *, int *);
extern void embed_config(bd_t **bp, unsigned long load_addr);

#ifdef CONFIG_SILKWORM

/* SIN (system inventory) function vectors into the bootprom */
/* see the prom vob for more details */
#if defined(CONFIG_SILKWORM440_AP)
#define SIN_REWIND_VECTOR	0xfff00b10
#define SIN_GET_VECTOR		0xfff00b14
#define SIN_STATS_VECTOR	0xfff00b18
#elif defined(CONFIG_SILKWORM440) || defined(CONFIG_SILKWORM440GX)
#define SIN_REWIND_VECTOR	0xfffa0b10
#define SIN_GET_VECTOR		0xfffa0b14
#define SIN_STATS_VECTOR	0xfffa0b18
#else	/* !CONFIG_SILKWORM440 */
#define SIN_REWIND_VECTOR	0xfffe0be0
#define SIN_GET_VECTOR		0xfffe0be4
#define SIN_STATS_VECTOR	0xfffe0be8
#endif /* CONFIG_SILKWORM440 */

typedef int (*sin_rewind_pt)(int *statep);
typedef sin_rec_t * (*sin_get_pt)(sin_rec_t *sr, int *statep);
typedef int (*sin_stats_pt)(sin_stats_t *ssp);

#define SIN_FUNCTIONS_NOT_DEFINED	-1
#define SIN_OUT_OFMEMORY		-2

/* Rudimentry allocation, derived from zalloc in boot/common/misc-common.c.
   Allocates and returns the pointer to requested memory
   Depends on avail_ram and end_avail pointers being correctly set 
*/
static void *sinalloc(unsigned int size)
{
	void *p = avail_ram;
	
	size = (size + 7) & -8;
	avail_ram += size;
	if (avail_ram > end_avail) {
		p = NULL;
	}
	return p;
}

/* Used to read sin_records and return these in a single table
   Uses sin_stats to get record size and number, allocates the needed memory
   and copies each record into the table.
    */
static int read_sin_records(void** sin_recs, int* sin_recs_size, unsigned long entry_point)
{
    /* The function pointers to the actual sin functions in the boot prom */
    sin_rewind_pt sin_rewind_p;
    sin_get_pt sin_get_p;
    sin_stats_pt sin_stats_p;
    sin_stats_t sin_table_stats;
    sin_rec_t *srp = NULL;
    int state, nsr=0;
    entry_point_struct *eps = (entry_point_struct *)entry_point;


    /* Initialize the sin function pointers */
    if (eps->magic1 == MAGIC1 && eps->magic2 == MAGIC2 && eps->version > MIN_VERSION &&
        eps->sin_rewind_vector != NULL && eps->sin_get_vector != NULL &&
        eps->sin_stats_vector != NULL)
    {
        sin_rewind_p = (void *)eps->sin_rewind_vector;
        sin_get_p = (void *)eps->sin_get_vector;
        sin_stats_p = (void *)eps->sin_stats_vector;
    }
    else
    {
        sin_rewind_p = (void *)(*(unsigned long *)SIN_REWIND_VECTOR);
        sin_get_p = (void *)(*(unsigned long *)SIN_GET_VECTOR);
        sin_stats_p = (void *)(*(unsigned long *)SIN_STATS_VECTOR);
    }



    /* Check for null as in older proms, sin_rewind is not defined */
    if(!sin_rewind_p)
	return SIN_FUNCTIONS_NOT_DEFINED;

    sin_stats_p(&sin_table_stats);

    /* Just printing a message and continuing, though there is a 
       chance that the sin records might not be accurate.
    */
    if(SIN_VERSION != sin_table_stats.ss_version)
	_printk("read_sin_records: SIN versions do not match, SIN_VERSION(%d) "\
	    "!= sin_table_stats.ss_version (%d), please update kernel to "\
	    "match bootprom.\n", SIN_VERSION, sin_table_stats.ss_version);

    /* Total size = recs * sizeof recs */
    *sin_recs_size = sin_table_stats.ss_used * sin_table_stats.ss_size;

    *sin_recs = sinalloc(*sin_recs_size);
    if(!*sin_recs)
    {
	*sin_recs_size=0;
	return SIN_OUT_OFMEMORY;
    }

    srp = (sin_rec_t *) *sin_recs;

    sin_rewind_p(&state);

    /* Copy all the sin records */
    while ( ((srp = sin_get_p(srp, &state)) != NULL) 
	      && (nsr < sin_table_stats.ss_used)) {
	srp++;
	nsr++;
    } /* while */

    return 0;
}

#endif /* CONFIG_SILKWORM */



unsigned long
load_kernel(unsigned long load_addr, int num_words, unsigned long cksum, bd_t *bp)
{
	char *cp, ch;
	int timer = 0, zimage_size;
	unsigned long initrd_size;
	void *sin_recs = NULL;	/* system inventory */
	int sin_recs_size;

	/* First, capture the embedded board information.  Then
	 * initialize the serial console port.
	 */
	embed_config(&bp, load_addr);
#if defined(CONFIG_SERIAL_CPM_CONSOLE) || defined(CONFIG_SERIAL_8250_CONSOLE)
	com_port = serial_init(0, bp);
#endif
	/* Grab some space for the command line and board info.  Since
	 * we no longer use the ELF header, but it was loaded, grab
	 * that space.
	 */
#ifdef CONFIG_MBX
	/* Because of the way the MBX loads the ELF image, we can't
	 * tell where we started.  We read a magic variable from the NVRAM
	 * that gives us the intermediate buffer load address.
	 */
	load_addr = *(uint *)0xfa000020;
	load_addr += 0x10000;		/* Skip ELF header */
#endif
	puts("\nloading kernel\n");

	/* copy board data */
	if (bp)
		memcpy(hold_residual,bp,sizeof(bd_t));

	/* Set end of memory available to us.  It is always the highest
	 * memory address provided by the board information.
	 */
	end_avail = (char *)(bp->bi_memsize);

	puts("\nloaded at:     "); puthex(load_addr);
	puts(" "); puthex((unsigned long)(load_addr + (4*num_words))); puts("\n");
	if ( (unsigned long)load_addr != (unsigned long)&start ) {
		puts("relocated to:  "); puthex((unsigned long)&start);
		puts(" ");
		puthex((unsigned long)((unsigned long)&start + (4*num_words)));
		puts("\n");
	}

	if ( bp ) {
		puts("board data at: "); puthex((unsigned long)bp);
		puts(" ");
		puthex((unsigned long)((unsigned long)bp + sizeof(bd_t)));
		puts("\nrelocated to:  ");
		puthex((unsigned long)hold_residual);
		puts(" ");
		puthex((unsigned long)((unsigned long)hold_residual + sizeof(bd_t)));
		puts("\n");
	}

	/*
	 * We link ourself to an arbitrary low address.  When we run, we
	 * relocate outself to that address.  __image_being points to
	 * the part of the image where the zImage is. -- Tom
	 */
	zimage_start = (char *)(unsigned long)(&__image_begin);
	zimage_size = (unsigned long)(&__image_end) -
			(unsigned long)(&__image_begin);

	initrd_size = (unsigned long)(&__ramdisk_end) -
		(unsigned long)(&__ramdisk_begin);

	/*
	 * The zImage and initrd will be between start and _end, so they've
	 * already been moved once.  We're good to go now. -- Tom
	 */
	puts("zimage at:     "); puthex((unsigned long)zimage_start);
	puts(" "); puthex((unsigned long)(zimage_size+zimage_start));
	puts("\n");

	if ( initrd_size ) {
		puts("initrd at:     ");
		puthex((unsigned long)(&__ramdisk_begin));
		puts(" "); puthex((unsigned long)(&__ramdisk_end));puts("\n");
	}

	/*
	 * setup avail_ram - this is the first part of ram usable
	 * by the uncompress code.  Anything after this program in RAM
	 * is now fair game. -- Tom
	 */
	avail_ram = (char *)PAGE_ALIGN((unsigned long)_end);

	puts("avail ram:     "); puthex((unsigned long)avail_ram); puts(" ");
	puthex((unsigned long)end_avail); puts("\n");

#ifdef CONFIG_SILKWORM
	/* This call has to be after setting avail_ram and before the kernel
	 * is copied and unzipped.  Once the kernel is copied/unzipped, 
	 * portions of the prom have been overwritten.
	 */
	read_sin_records(&sin_recs, &sin_recs_size, load_addr);
#endif

	puts("\nLinux/PPC load: ");
	cp = cmd_line;
	/* This is where we try and pick the right command line for booting.
	 * If we were given one at compile time, use it.  It Is Right.
	 * If we weren't, see if we have a ramdisk.  If so, thats root.
	 * When in doubt, give them the netroot (root=/dev/nfs rw) -- Tom
	 */
#ifdef CONFIG_CMDLINE_BOOL
	memcpy (cmd_line, compiled_string, sizeof(compiled_string));
#else
#ifdef CONFIG_SILKWORM
	/* If no explicit cmdline, then don't use anything. */
	*cmd_line = '\0';
#else
	if ( initrd_size )
		memcpy (cmd_line, ramroot_string, sizeof(ramroot_string));
	else
		memcpy (cmd_line, netroot_string, sizeof(netroot_string));
#endif /* CONFIG_SILKWORM */
#endif
	if(*bootrom_cmdline != '\0') {
		puts("\nBootROM command line: ");
		cp = bootrom_cmdline;
		memcpy(treeboot_bootrom_cmdline, bootrom_cmdline, 
			strlen(bootrom_cmdline)+1);
		cmd_line = treeboot_bootrom_cmdline;
		cp = cmd_line;
	}
	while ( *cp )
		putc(*cp++);
/*
 * Skip kernel\bootloader command line modification unless it is for debugging
 * Saves seconds from bootup time and avoids a security hole.
 */
#ifdef CONFIG_SERIAL_TEXT_DEBUG
#if defined(CONFIG_SERIAL_CONSOLE) || defined(CONFIG_VGA_CONSOLE)
	/*
	 * If they have a console, allow them to edit the command line.
	 * Otherwise, don't bother wasting the five seconds.
	 */
	while (timer++ < 5*1000) {
		if (tstc()) {
			while ((ch = getc()) != '\n' && ch != '\r') {
				if (ch == '\b' || ch == '\177') {
					if (cp != cmd_line) {
						cp--;
						puts("\b \b");
					}
				} else if (ch == '\030'		/* ^x */
					   || ch == '\025') {	/* ^u */
					while (cp != cmd_line) {
						cp--;
						puts("\b \b");
					}
				} else {
					*cp++ = ch;
					putc(ch);
				}
			}
			break;  /* Exit 'timer' loop */
		}
		udelay(1000);  /* 1 msec */
	}
#endif
#endif /* ifdef CONFIG_SERIAL_TEXT_DEBUG */

	*cp = 0;
	puts("\nUncompressing Linux...");

	gunzip(0, 0x600000, zimage_start, &zimage_size);
	flush_instruction_cache();
	puts("done.\n");
	{
		struct bi_record *rec;
		unsigned long initrd_loc = 0;
		unsigned long rec_loc = _ALIGN((unsigned long)(zimage_size) +
				(1 << 20) - 1, (1 << 20));
		rec = (struct bi_record *)rec_loc;

		/* We need to make sure that the initrd and bi_recs do not
		 * overlap. */
		if ( initrd_size ) {
			initrd_loc = (unsigned long)(&__ramdisk_begin);
			/* If the bi_recs are in the middle of the current
			 * initrd, move the initrd to the next MB
			 * boundary. */
			if ((rec_loc > initrd_loc) &&
					((initrd_loc + initrd_size)
					 > rec_loc)) {
				initrd_loc = _ALIGN((unsigned long)(zimage_size)
						+ (2 << 20) - 1, (2 << 20));
			 	memmove((void *)initrd_loc, &__ramdisk_begin,
					 initrd_size);
		         	puts("initrd moved:  "); puthex(initrd_loc);
			 	puts(" "); puthex(initrd_loc + initrd_size);
			 	puts("\n");
			}
		}

		rec->tag = BI_FIRST;
		rec->size = sizeof(struct bi_record);
		rec = (struct bi_record *)((unsigned long)rec + rec->size);

		rec->tag = BI_CMD_LINE;
		memcpy( (char *)rec->data, cmd_line, strlen(cmd_line)+1);
		rec->size = sizeof(struct bi_record) + strlen(cmd_line) + 1;
		rec = (struct bi_record *)((unsigned long)rec + rec->size);

		if ( initrd_size ) {
			rec->tag = BI_INITRD;
			rec->data[0] = initrd_loc;
			rec->data[1] = initrd_size;
			rec->size = sizeof(struct bi_record) + 2 *
				sizeof(unsigned long);
			rec = (struct bi_record *)((unsigned long)rec +
					rec->size);
		}
#ifdef CONFIG_SILKWORM
		if ( sin_recs ) {
			rec->tag = BI_SILKWORM_SIN_TBL;
			memcpy( (char *)rec->data, sin_recs, sin_recs_size);
			rec->size = sizeof(struct bi_record) + sin_recs_size;
			rec = (struct bi_record *)((unsigned long)rec + 
					rec->size);
		}
#endif
		rec->tag = BI_LAST;
		rec->size = sizeof(struct bi_record);
		rec = (struct bi_record *)((unsigned long)rec + rec->size);
	}
	puts("Now booting the kernel\n");
#if defined(CONFIG_SERIAL_CPM_CONSOLE) || defined(CONFIG_SERIAL_8250_CONSOLE)
	serial_close(com_port);
#endif

	return (unsigned long)hold_residual;
}
