/*
 * BK Id: %F% %I% %G% %U% %#%
 * 
 * Makes a tree bootable image for IBM Evaluation boards.
 * Basically, just take a zImage, skip the ELF header, and stuff
 * a 32 byte header on the front.
 *
 * We use htonl, which is a network macro, to make sure we're doing
 * The Right Thing on an LE machine.  It's non-obvious, but it should
 * work on anything BSD'ish.
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netinet/in.h>
#include <elf.h>

/* This gets tacked on the front of the image.  There are also a few
 * bytes allocated after the _start label used by the boot rom (see
 * head.S for details).
 */
typedef struct boot_block {
	unsigned long bb_magic;		/* 0x0052504F */
	unsigned long bb_dest;		/* Target address of the image */
	unsigned long bb_num_512blocks;	/* Size, rounded-up, in 512 byte blks */
	unsigned long bb_debug_flag;	/* Run debugger or image after load */
	unsigned long bb_entry_point;	/* The image address to start */
	unsigned long bb_checksum;	/* 32 bit checksum including header */
	unsigned long reserved[2];
} boot_block_t;

#define MIN(a, b)	((a) < (b) ? (a) : (b))
#define IMGBLK	512
char	tmpbuf[IMGBLK];

int main(int argc, char *argv[])
{
	int	in_fd, out_fd;
	int	i, offset;
	unsigned long	nblks, cksum, *cp, nbytes, pad_bytes, readbytes;
	struct	stat	st;
	boot_block_t	bt;
	unsigned char padding[] = { 0xde, 0xad, 0xbe, 0xef };
	Elf32_Phdr *elf_ph;
	Elf32_Ehdr *elf_ex;

	if (argc < 3) {
		fprintf(stderr, "usage: %s <zImage-file> <boot-image> [entry-point]\n",argv[0]);
		exit(1);
	}

	if (stat(argv[1], &st) < 0) {
		perror("stat");
		exit(2);
	}
	nbytes = st.st_size;

	bt.bb_magic = htonl(0x0052504F);

	/* If we have the optional entry point parameter, use it */
	if (argc == 4)
		bt.bb_dest = bt.bb_entry_point = htonl(strtoul(argv[3], NULL, 0));
	else
		bt.bb_dest = bt.bb_entry_point = htonl(0x400000);

	/* We know these from the linker command.
	 * ...and then move it up into memory a little more so the
	 * relocation can happen.
	 */
	bt.bb_debug_flag = 0;
	bt.bb_checksum = 0;

	/* To be neat and tidy :-).
	*/
	bt.reserved[0] = 0;
	bt.reserved[1] = 0;

	if ((in_fd = open(argv[1], O_RDONLY)) < 0) {
		perror("zImage open");
		exit(3);
	}

	if ((out_fd = open(argv[2], (O_RDWR | O_CREAT | O_TRUNC), 0666)) < 0) {
		perror("bootfile open");
		exit(3);
	}

	/* Assume zImage is an ELF file, and skip the 64K header.
	*/
	if (read(in_fd, tmpbuf, IMGBLK) != IMGBLK) {
		fprintf(stderr, "%s is too small to be an ELF image\n",
				argv[1]);
		exit(4);
	}

	if ((*(uint *)tmpbuf) != htonl(0x7f454c46)) {
		fprintf(stderr, "%s is not an ELF image\n", argv[1]);
		exit(4);
	}

	/* Determine the offset to our first section */
	elf_ex = (Elf32_Ehdr*) tmpbuf;
	if (lseek(in_fd, ntohl(elf_ex->e_phoff), SEEK_SET) < 0) {
		fprintf(stderr, "%s failed to seek in ELF image\n", argv[1]);
		exit(4);
	}
	if (read(in_fd, tmpbuf, IMGBLK) != IMGBLK) {
		fprintf(stderr, "%s is too small to be an ELF image\n",
				argv[1]);
		exit(4);
	}
	elf_ph = (Elf32_Phdr*) tmpbuf;
	offset = ntohl(elf_ph->p_offset);
	if(offset % 4096) {
		fprintf(stderr, "%s seems to have a bad offset [0x%x] for the program segment\n",
				argv[1], offset);
		exit(4);
	}
	if (lseek(in_fd, offset, SEEK_SET) < 0) {
		fprintf(stderr, "%s failed to seek in ELF image\n", argv[1]);
		exit(4);
	}
	nbytes -= offset;

	nblks = (nbytes + sizeof(bt) + IMGBLK - 1) / IMGBLK;
	pad_bytes = (nblks * IMGBLK) - nbytes - sizeof(bt);
	bt.bb_num_512blocks = htonl(nblks);

	/* And away we go......
	*/
	if (write(out_fd, &bt, sizeof(bt)) != sizeof(bt)) {
		perror("boot-image write header");
		exit(5);
	}

	while (nbytes > 0) {
		readbytes = MIN(IMGBLK, nbytes);
		nbytes -= readbytes;
		if (read(in_fd, tmpbuf, readbytes) < 0) {
			perror("zImage read");
			exit(5);
		}
		if (write(out_fd, tmpbuf, readbytes) != readbytes) {
			perror("boot-image write image");
			exit(5);
		}
	}

	for(i=0; i<sizeof(tmpbuf); i++)
		tmpbuf[i] = padding[i%sizeof(padding)];

	if (write(out_fd, tmpbuf, pad_bytes) != pad_bytes) {
		perror("boot-image write padding");
		exit(5);
	}

	/* compute checksum over entire file 
	*/
	close(out_fd);
	if ((out_fd = open(argv[2], (O_RDWR), 0666)) < 0) {
		perror("bootfile open");
		exit(3);
	}
	cksum = 0;
	nbytes = nblks * IMGBLK;
	while (nbytes > 0) {
		if (read(out_fd, tmpbuf, IMGBLK) != IMGBLK) {
			perror("zImage read");
			exit(5);
		}
		nbytes -= IMGBLK;
		cp = (unsigned long*)tmpbuf;
		for (i=0; i<IMGBLK/sizeof(unsigned long); i++)
		{
			unsigned long value = *cp++;
			cksum += ntohl(value);
		}
	}

	/* rewrite the header with the computed checksum.
	*/
	cksum = ~(cksum - 1);	// invert;
	bt.bb_checksum = htonl(cksum);
	if (lseek(out_fd, 0, SEEK_SET) < 0) {
		perror("rewrite seek");
		exit(1);
	}
	if (write(out_fd, &bt, sizeof(bt)) != sizeof(bt)) {
		perror("boot-image rewrite");
		exit(1);
	}
	close(out_fd);
	close(in_fd);

	fprintf(stderr, "Entry Point    : 0x%08lx\n", ntohl(bt.bb_entry_point));
	fprintf(stderr, "Image size     : 0x%08lx (%lu blocks)\n", 
			ntohl(bt.bb_num_512blocks) * IMGBLK, 
			ntohl(bt.bb_num_512blocks));
	fprintf(stderr, "ELF offset     : %d bytes\n", offset);
	fprintf(stderr, "Padding        : %lu bytes\n", pad_bytes);
	fprintf(stderr, "Checksum       : 0x%08lx\n", ntohl(bt.bb_checksum));
	exit(0);
}
