/*
 *
 *    Copyright (c) 1996-2009 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *    Description:
 *      Access routines for the CPLD on Brocade Silkworm switches
 *      Currently supported only for the 440 platforms, this can
 *	merged with silkworm_cpld.c - keeping it seperate for i2c virtual bus. 
 * 	related changes
 *
 */

#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/i2c.h>
#include <linux/i2c-virtual.h>

#include <asm/irq.h>
#include <asm/io.h>
#include <brocade/cpld.h>
#include <brocade/platform.h>
#include <brocade/silkworm.h>
#include "silkworm_cpld.h"

extern void kick_watchdog(void );

extern sw_bd silkworm;

#ifdef SILKWORM_DEBUG
#define DEBUG(x...)  printk(x)
#else
#define DEBUG(x...)
#endif

/* Our private data */
struct cpld_priv_t {
	phys_addr_t		paddr;
	char			*vaddr;
	int			max_size;
	int			initialized;
};
static struct cpld_priv_t	cpld_priv;
static DECLARE_MUTEX(i2c_sem);

static char *pci_ras = NULL;

/*
 * int cpld_init()
 *
 * Description:
 *   Map in the CPLD base address and initialize
 *   pointers to the registers to be used by other
 *   routines in this file.
 *
 *   For dual processor systems, init the I2C mux
 *   semaphore.
 *
 * Input(s):
 *   N/A
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *    0 = success
 *   -1 = fail
 *
 */
int
cpld_init(void)
{
	uint8_t		reg;

#if	0
retry:
#endif
	switch (silkworm.board) {
#if 0
		case 0:			/* legacy support for CP03 */
			printk("%s: Assuming legacy CP03 hardware\n", __FUNCTION__);
			silkworm.board = SWBD10;
			silkworm.board_rev = 1;
			goto retry;
			break;
		case SWBD10:
		case SWBD12:
			//printk("Silkworm CPLD - v1.1\n"); 
			cpld_priv.max_size 	= CPLD_MAX_REGISTERS;
			cpld_priv.paddr 	= CPLD_BASE_ADDRESS;
			break;
#endif
		case SWBD21:
		case SWBD29:
		case SWBD42:
		case SWBD46:
		case SWBD37:
		case SWBD43:
		case SWBD45:
		case SWBD51:
		case SWBD61:
#ifdef CONFIG_440 /* Real Meteor */
			//printk("Silkworm128 CPLD - v1.0\n");
			cpld_priv.max_size 	= CPLD_MAX_REGISTERS;
			cpld_priv.paddr 	= MET_MGMT_BASE_ADDR;
#else
			//printk("Silkworm CPLD - v1.1\n"); 
			cpld_priv.max_size 	= CPLD_MAX_REGISTERS;
			cpld_priv.paddr 	= CPLD_BASE_ADDRESS;
#endif
			break;
		case SWBD55:
		case SWBD62:
		case SWBD63:
		case SWBD67:
		case SWBD68:
		case SWBD76: /* ELARA */
		case SWBD23:
		case SWBD64:
		case SWBD70:
		case SWBD71:
		case SWBD72:
		case SWBD82:
		case SWBD66:
		case SWBD73:
		case SWBD75:
		case SWBD77:
		case SWBD79: /* Spear CP */
		case SWBD85: /* Spike-BP */
		case SWBD83: /* Spike */
		case SWBD88: /* Europa */
		case SWBD89: /* Striker */
		case SWBD86:
		case SWBD87:
		case SWBD92:
		case SWBD109: /* Stinger */
		case SWBD117: /* Pharos */
		case SWBD118:
		case SWBD124: /* Kestrel2 */
		case SWBD125:
		case SWBD129: /* Bulova2 */
		case SWBD130: /* Superhawk3 */
		case SWBD133:
		case SWBD134: /* Qilin */
		case SWBD149: /* Kestrel3 */
		case SWBD150: /* Falcon3 */
			cpld_priv.max_size 	= PULSAR_CPLD_SIZE;
			cpld_priv.paddr 	= PULSAR_CPLD_BASE_ADDR;
			break;
		case SWBD24:
			cpld_priv.max_size 	= ATHENA_CPLD_SIZE;
			cpld_priv.paddr 	= ATHENA_CPLD_BASE_ADDR;
			break;
		case SWBD53:
			cpld_priv.max_size 	= CPLD_MAX_REGISTERS;
			cpld_priv.paddr 	= PULSAR_CPLD_BASE_ADDR;
			break;

		default:
			printk("%s: platform (%d) not supported\n", 
				__FUNCTION__, silkworm.board);
			return -1;
			break;
	}

	/* remap the CPLD into our memory space */
	cpld_priv.vaddr = ioremap(cpld_priv.paddr, cpld_priv.max_size);
	if (!cpld_priv.vaddr) {
		printk("%s: Failed to map cpld\n", __FUNCTION__);
		return -1;
	}
	cpld_priv.initialized = 1;

	if ((silkworm.board == SWBD62)||(silkworm.board == SWBD77)) {

		/* Map in the PCI RAS PLD */

		printk(KERN_ERR "%s: Map PCI RAS device\n", __FUNCTION__);		
		pci_ras = ioremap(PCI_RAS_PLD_BASE_ADDR,  PCI_RAS_PLD_SIZE);
		if (!pci_ras) {
			printk(KERN_ERR "%s: Failed to map PCI RAS device\n",
				   __FUNCTION__);
		};

	}

	/* Some initialization that linux has to do */
	switch(silkworm.board) {

		case SWBD10:
			/* Clear this CP error bit */
			CPLD_CLEAR_MASK(CPLD_CP_ERROR_OUT, 
				(cpld_priv.vaddr + MISC_WRITE_OFFSET));
			break;

		case SWBD21:
		case SWBD42:
		case SWBD46:
		/* knj - SWBD36 does not need to clear masks */ 
		/* knj - do swbd36 specific initialization here */
		case SWBD55:
        case SWBD62:
		case SWBD64:
		case SWBD68:
        case SWBD77:
        case SWBD89:
		case SWBD133:
#if defined(CONFIG_440) || defined(CONFIG_85xx) /* Real Meteor */
			/* Clear this CP error bit */
			CPLD_CLEAR_MASK(MET_CP_SECTION_ERR_OUT, 
				(cpld_priv.vaddr + MET_CP_MISC_WRITE));
			CPLD_CLEAR_MASK(MET_HK_SELF_FENCE,
				(cpld_priv.vaddr + MET_CP_HK_FENCE));
			CPLD_CLEAR_MASK(CP_I2C_WR_ACTIVE_MASK,
				(cpld_priv.vaddr + MET_I2C_SELECT));
				
	
#else
			/* Clear this CP error bit */
			CPLD_CLEAR_MASK(CPLD_CP_ERROR_OUT, 
				(cpld_priv.vaddr + MISC_WRITE_OFFSET));
#endif
			break;
		case SWBD24:
			/* Unfence the global I2C bus, so we can use it */
			cpld_read_register(ATHENA_PROC_BLADE_FENCE_CNTRL_STATUS_REG, &reg);
			reg &= ~ATHENA_HK_SELF_FENCE;
			cpld_write_register(ATHENA_PROC_BLADE_FENCE_CNTRL_STATUS_REG, reg);

			/* During board bringup we have to fake out the 
			 * optical switch to make it _think_ the board is 
			 * properly seated in the chassis. This is done with 
			 * reflective tape.
			 */
			cpld_read_register(ATHENA_LCP_SLOT_ID_SEATED_REG, &reg);
			if((reg & ATHENA_BOARD_SEATED))
				printk("\n\n*** WARNING: The BOARD_SEATED_L "
				"signal is not correct!\n"
				"Is your board missing reflective tape on "
				"the optical sensor?"
				"\n\n");
			
			break;
	}
	printk("Silkworm CPLD [board: %d, rev: %d]\n", silkworm.board, 
		silkworm.board_rev);
	return (0);
}


/*
 * int cpld_read_register()
 *
 * Description:
 *   Provides read access to CPLD registers.
 *
 * Input(s):
 *   offset - register offset from the CPLD base address
 *   *data  - Storage to contain CPLD register contents
 *
 * Output(s):
 *   Contents of specified CPLD register.
 *
 * Returns:
 *   Success - 0
 *   Fail    - < 0
 *
 */
int
cpld_read_register(int offset, unsigned char *data)
{
	if (!cpld_priv.initialized) {
	  printk("cpld_read_register: Cannot access CPLD. "
		 "Initialization needed\n");
	  return (-1);
	}

	if (!data) {
	  printk(KERN_ERR "cpld_read_register: "
		 "Uninitialized input parameter\n");
	  return (-1);
	}

	if (offset >= cpld_priv.max_size) {
	  printk(KERN_ERR "cpld_read_register: "
		 "Specified register is out of range: %d  max allowed: %d\n",
		 offset, CPLD_MAX_REGISTERS);
	  return (-1);
	}

	*data = readb(cpld_priv.vaddr + offset);

	return (0);
}

/*
 * int cpld_write_register()
 *
 * Description:
 *   Provides write access to CPLD registers.
 *
 * Input(s):
 *   offset - register offset from the CPLD base address
 *   data   - Data to write to the CPLD register
 *
 * Output(s):
 *   Contents of specified CPLD register.
 *
 * Returns:
 *   Success - 0
 *   Fail    - < 0
 *
 */
int
cpld_write_register(int offset, unsigned char data)
{
	if (!cpld_priv.initialized) {
	  printk("cpld_write_register: Cannot access CPLD. "
		 "Initialization needed\n");
	  return (-1);
	}

	if (offset >= cpld_priv.max_size) {
	  printk(KERN_ERR "cpld_write_register: "
		 "Specified register is out of range: %d  max allowed: %d\n",
		 offset, CPLD_MAX_REGISTERS);
	  return (-1);
	}

	writeb(data, cpld_priv.vaddr + offset);
	return 0;
}

/* Returns slot id for this cp, usually slot 5, or slot 6.  Returns 0 for non-bladed
 * chassis.
 */
int
cpld_get_slotid(void)
{
	uint8_t	reg;
	int data = 0;

	switch(silkworm.board) {
		case SWBD10:
#if !defined(CONFIG_440) && !defined(CONFIG_85xx)
		case SWBD21:	/* fake meteor = Ulysses */
#endif
			cpld_read_register(MISC_READ_OFFSET, &reg);
			if(reg & CPLD_SLOT_ID_FIELD)
				data = CPLD_CP1_I2C_ADDRESS;
			else
				data = CPLD_CP0_I2C_ADDRESS;
			break;
#if defined(CONFIG_440) || defined(CONFIG_85xx) 	/* real meteor */
		case SWBD55: /* Scimitar Freescale based */
		case SWBD62:
		case SWBD68:
		case SWBD21:
		case SWBD42:
		case SWBD64:
		case SWBD77:
		case SWBD89:
		case SWBD133:
			cpld_read_register(MET_CP_MISC_READ, &reg);
			if(reg & MET_CP_CP_SLOT_ID)
				data = MET_CPLD_CP1_I2C_ADDR;
			else
				data = MET_CPLD_CP0_I2C_ADDR;
			break;
#endif
		case SWBD24:	/* Olympus */
			cpld_read_register(ATHENA_LCP_SLOT_ID_SEATED_REG, &reg);
			data &= ATHENA_SLOT_ID_MASK;
			break;
			
		default:
			data = 0;
			printk("%s: unknown platform %d\n", 
				__FUNCTION__, silkworm.board);
	}
	return data;
}


EXPORT_SYMBOL(cpld_read_register);
EXPORT_SYMBOL(cpld_write_register);
EXPORT_SYMBOL(cpld_get_slotid);


static int sem_taken = 0;
/* This is a hack for now, to support this legacy method of accessing the 
 * shared i2c bus. We know the i2c driver has code in place to acquire/release 
 * exclusive access to the bus. Use these same functions to do the dirty work.
 * cpld_release_i2c_mux fn also has to set the mux to a deselected position.
 */
int cpld_arbitrate_i2c_mux(void) 
{
	int status;
	int (*acquire_exclusive)(struct i2c_adapter *adap);
	struct i2c_adapter *adap = i2c_get_adapter(1);	/* iic1 is shared bus */

	/* FIXME!! */
	/* Hack Alert!  Take a semaphore here, as the caller expects this 
	* function to * not only obtain exclusive bus access w.r.t. the other 
	* CP, but also w.r.t.  * any other processes on this CP.  This will 
	* cause problems if we have * code using both the new and the old 
	* systems of i2c access (sensor drivers * use new method, FabOS still 
	* uses old method), as the new mechanism doesn't * use this function, 
	* and therefore wouldn't respect the semaphore lock.  * * That being 
	* said, currently the only linux I2C client is the RTC, which is * on 
	* the iic0 (local) bus.  It doesn't call this function, but it also * 
	* doesn't use a MUX, so for now this is safe...  * * This whole 
	* function will go away as soon as FabOS is updated to use the new * 
	* virtual adapters.
	 */
	down(&i2c_sem);
	if (sem_taken == 1) {
	  printk("\ncpld_arbitrate_i2c_mux WARNING: I2C LOCAL "
		 "SEMAPHORE FAILURE.");
	}
	sem_taken = 1;

	if(!adap)
		return 0;	/* this platform doesn't have one */

	acquire_exclusive = adap->algo->acquire_exclusive;
	if (!acquire_exclusive)
		return 0;

	/* acquire_exclusive is called w/o holding the I2C_LOCK, so this is ok */
	status = acquire_exclusive(adap);

	if (status < 0) {
		sem_taken = 0;
		up(&i2c_sem);
	}
	return (status);
}

int cpld_release_i2c_mux(int force) 
{ 
	int (*release_exclusive)(struct i2c_adapter *adap);
	struct i2c_adapter	*adap;
	struct i2c_virt_priv    *priv = NULL;
	struct i2c_mux_ctrl	mux;
	int			i, ret;

	/* FIXME - what should 'force' do?? */

	/* We assume at least one of the iic busses has a multiplexing function. 
	 * Just pick the first one we find.
	 */
	for(i=0; i<I2C_ADAP_MAX; i++) {
		if((adap = i2c_get_adapter(i)) && 
		  ((adap->id & 0xff) == I2C_HW_VIRT) &&
		   (priv = i2c_get_adapdata(adap)) &&
		    priv->mux.select != NULL) {
		    	break;
		}
	}

	if(adap == NULL || priv == NULL || priv->mux.select == NULL) {
		//printk("%s: didn't find a mux function\n", __FUNCTION__);
		ret = 0;
	} else {
		mux.addr = 0;		/* the legacy platforms (ulysses, term) don't use addr */
		mux.value = 0;		/* not needed for deselect */
		mux.select = priv->mux.select;
		mux.deselect = priv->mux.deselect;

		/* This has to be called with the I2C_LOCK held, since it might have to perform
	 	 * an I2C transfer to select the mux.
	 	 */
		down(&adap->bus_lock);
		priv->mux.deselect(adap, &mux);
		up(&adap->bus_lock);

		/* release_exclusive is called w/o holding the I2C_LOCK, so this is ok */
		release_exclusive = adap->algo->release_exclusive;
		ret = ((release_exclusive) ? release_exclusive(adap) : 0);
	}

	/* Release the hackish global i2c semaphore */
	if (sem_taken) {
		sem_taken = 0;
		up(&i2c_sem);
	}
	return ret;
}

int cpld_set_i2c_mux(int use_slot_id, char data) 
{
	int i;
	struct i2c_adapter	*adap = NULL;
	struct i2c_virt_priv    *priv = NULL;
	struct i2c_mux_ctrl	mux;
	
	/* First handle this use_slot_id crap */
	if(use_slot_id) 
		data = cpld_get_slotid();

	/* We assume at least one of the iic busses has a multiplexing function. 
	 * Just pick the first one we find.
	 */
	for(i=0; i<I2C_ADAP_MAX; i++) {
		if((adap = i2c_get_adapter(i)) && 
		  ((adap->id & 0xff) == I2C_HW_VIRT) &&
		   (priv = i2c_get_adapdata(adap)) &&
		    priv->mux.select != NULL) {
		    	break;
		}
	}
	if(adap == NULL || priv == NULL || priv->mux.select == NULL)
	{
		//printk("%s: didn't find a mux function\n", __FUNCTION__);
		return 0;
	}

	mux.addr = 0;		/* the legacy platforms (ulysses, term) don't use addr */
	mux.value = data;
	mux.select = priv->mux.select;
	mux.deselect = priv->mux.deselect;

	/* Get exclusive access to the shared bus */
	if(cpld_arbitrate_i2c_mux() != 0) {
		printk("%s: Failed to get exclusive access to I2C bus," 
			"mux setting %d\n",
			__FUNCTION__, data);
		return -EIO;
	}

	/* This has to be called with the I2C_LOCK held, since it might have 
	 * to perform
	 * an I2C transfer to select the mux.
	 */
	down(&adap->bus_lock);
	priv->mux.select(adap, &mux);
	up(&adap->bus_lock);

	return 0;
}

int cpld_set_i2c_mux_local(void) 
{ 
	/* Select our own slow, using the use_slot_id flag */
	return cpld_set_i2c_mux(1, 0);
}
int cpld_get_i2c_mux(char *data) 
{ 
	*data = 0;
	return 0;
}

/* Here is the code for the CPLD SPI controller (used for management access
 * to the ethernet switch on Olympus).
 */
static
void cpld_decode_spi_err(int val) 
{
	/* Process error bits */
	if(val & ATHENA_SPI_CMD_ERR)
		printk("SPI read command error!\n");
	if(val & ATHENA_SPI_INVALID_CMD)
		printk("SPI read command invalid!\n");
	if(val & ATHENA_SPI_SPIF_TIMEOUT)
		printk("SPIF bit not set\n");
	if(val & ATHENA_SPI_RDY_TIMEOUT)
		printk("SPI ready bit not set\n");
	if(val & ATHENA_SPI_RACK_TIMEOUT)
		printk("SPI ACK timneout\n");
}

/*
 * cpld_spi_Read():  Read Broadcom switch register thru SPI.
 */
int cpld_spi_read(uint8_t unit, uint8_t page, uint8_t offset, uint8_t *data,
		int datalen)
{
	unsigned long start;
	int i;
	uint8_t val;

	if (!data)
		return -EFAULT;

	if (unit >= 2)	/* only two SPI chip selects */
		return -EINVAL;

	if (datalen > ATHENA_MAX_SPI_DATA_LEN) {
		/* Will support when have hardware to verify code */
		printk("%s: SPI driver currently is not supporting >%d bytes read operation.\n",
			__FUNCTION__, ATHENA_MAX_SPI_DATA_LEN);
		return -E2BIG;
	}

	cpld_read_register(ATHENA_SPI_STATUS_REG, &val);
	if ((val & ATHENA_SPI_CMD_DONE) == 0) {
		printk("ERROR: Previous SPI read/write command was not done!\n");
		return -EIO;
	}
	
	/* Assert chip select to select BCM device */
	cpld_read_register(ATHENA_SPI_SWITCH_SEL_CNTRL_REG, &val);
	if (unit == 1)
		val |= ATHENA_SPI_SW_SEL;
	else
		val &= ~ATHENA_SPI_SW_SEL;
	cpld_write_register(ATHENA_SPI_SWITCH_SEL_CNTRL_REG, val);

	/* Setup page and data address */
	cpld_write_register(ATHENA_SPI_PAGE_ADDR_REG, page);
	cpld_write_register(ATHENA_SPI_OFFSET_ADDR_REG, offset);

	/* Prepare control register for read op */
	val = (datalen << 4) | ATHENA_SPI_RD_WR_1 | ATHENA_SPI_EN;
	cpld_write_register(ATHENA_SPI_CNTRL_REG, val);

	/* Start the read command */
	val |= ATHENA_SPI_CMD_VALID;
	cpld_write_register(ATHENA_SPI_CNTRL_REG, val);    

	/* Poll for done bit */
	start = jiffies;
	do {
		cpld_read_register(ATHENA_SPI_STATUS_REG, &val);
		if (val & ATHENA_SPI_CMD_DONE)
			break;
		schedule_timeout(1);
	} while (jiffies < (start + HZ));
	
	if (!(val & ATHENA_SPI_CMD_DONE)) {
		printk("%s: SPI read timeout! \n", __FUNCTION__);
		return -EIO;
	}

	val &= ATHENA_SPI_ERR_MASK;
	if (val != 0) {
		cpld_write_register(ATHENA_SPI_STATUS_REG, val);    
		cpld_decode_spi_err(val);
		return -EIO;
	} else {
		/* Move data to buffer */
		for (i=0; i<datalen; i++)
			cpld_read_register(ATHENA_SPI_DATA_BYTE0_REG + i, &data[i]);
	}

	return 0;
}

/*--------------------------------------------------------------------------+
 | cpld_spi_Write: Write BCM register thru SPI.
 +---------------------------------------------------------------------------*/
int cpld_spi_write(uint8_t unit, uint8_t page, uint8_t offset, uint8_t *data,
                int datalen)
{
	unsigned long start;
	int i;
	uint8_t val;

	if (!data)
		return -EFAULT;
	
	if (unit >= 2)		/* only 2 spi chip selects */
		return -EINVAL;

	if (datalen > ATHENA_MAX_SPI_DATA_LEN) {
		/* Will support when have hardware to verify code */
		printk("%s: SPI driver currently is not supporting >%d bytes read operation.\n",
			__FUNCTION__, ATHENA_MAX_SPI_DATA_LEN);
		return -E2BIG;
	}

	cpld_read_register(ATHENA_SPI_STATUS_REG, &val);
	if ((val & ATHENA_SPI_CMD_DONE) == 0) {
		printk("%s: Previous SPI read/write command was not done!\n", __FUNCTION__);
		return -EIO;
	}

	/* Select BCM device */
	cpld_read_register(ATHENA_SPI_SWITCH_SEL_CNTRL_REG, &val);
	if (unit == 1)
		val |= ATHENA_SPI_SW_SEL;
	else
		val &= ~ATHENA_SPI_SW_SEL;
	cpld_write_register(ATHENA_SPI_SWITCH_SEL_CNTRL_REG, val);

	/* Setup page and data address */
	cpld_write_register(ATHENA_SPI_PAGE_ADDR_REG, page);
	cpld_write_register(ATHENA_SPI_OFFSET_ADDR_REG, offset);
	
	/* Copy data to SPI data registers */		
	for (i=0; i<datalen; i++)
		cpld_write_register(ATHENA_SPI_DATA_BYTE0_REG+i, data[i]); 
	
	/* Prepare control register for write op */
	val = (datalen << 4) | ATHENA_SPI_EN;
	val &= ~ATHENA_SPI_RD_WR_1;
	cpld_write_register(ATHENA_SPI_CNTRL_REG, val);
	
	/* Start the write command */
	val |= ATHENA_SPI_CMD_VALID;
	cpld_write_register(ATHENA_SPI_CNTRL_REG, val);    /* Write a 1 to cmd_valid */
	
	/* Poll for done bit */
	start = jiffies;
	while (jiffies < (start + HZ)) {	/* wait 1 sec max */ 
		cpld_read_register(ATHENA_SPI_STATUS_REG, &val);
		if (val & ATHENA_SPI_CMD_DONE)
			break;
		schedule_timeout(1);
	}
	
	if (!(val & ATHENA_SPI_CMD_DONE)) {
		printk("%s: SPI write timeout!\n", __FUNCTION__);
		return -EIO;
	}
	
	val &= ATHENA_SPI_ERR_MASK;
	if (val != 0) {
		cpld_write_register(ATHENA_SPI_STATUS_REG, val);    
		cpld_decode_spi_err(val);
		return -EIO;
	}

	return 0;
}

/* other code needs access to the cpld base address */

void *cpld_get_base_addr(void)
{
	return ((void *) cpld_priv.vaddr);
}

/*
 * int cpld_check_pci_reset
 *
 * Description:
 *   See if a PCI bridge reset occured. If so, a PCI timeout has occured.
 *
 *   This applies only to systems that have the PCI RAS device.
 *
 *   It's essential that the HW watchdog not be allowed to expire while
 *   processing a PCI timeout.
 *
 * Input(s):
 *   None
 *
 * Output(s):
 *   None
 *
 * Returns:
 *   0 = No bridge reset or PCI timeout occured, or the system
 *       does not have the PCI RAS device.
 *   1 = Bridge reset or PCI timeout occured.
 */
int cpld_check_pci_reset(void)
{
	uint8_t val, val1, cbe = 0;
	uint8_t ad0 = 0, ad1 = 0, ad2 = 0, ad3 = 0;
	int address, ret = 0;


	if (!pci_ras) {
		ret = 0;
		goto exit;
	}

	kick_watchdog();

	cpld_read_register(CPLD_MISC_INTERRUPT, &val1);

	val = val1 & PCI_TIMEOUT_IRQ_BRIDGES;

	if (val) {
		int orig_log_level;		

		kick_watchdog();
		cbe = readb(pci_ras + PCI_TIMEOUT_CBE);

		/*
		 * Raise the log level to ensure any messages get to the console. In
		 * the event of a sudden reboot, this may be the only way we have of
		 * knowing what did or did not happen.
		 */

		orig_log_level = console_loglevel;
		console_verbose();

		printk(KERN_ERR "%s: PCI Timeout and bridge reset occured: 0x%x 0x%x cbe 0x%x\n",
			   __FUNCTION__, val, val1, cbe);

		kick_watchdog();
		console_loglevel = orig_log_level;
		ret = 1;
	}

 exit:
	return (ret);
}


/*
 * cpld_power_off()
 *
 * Description:
 *  Hardware specific power-down routine  for Brocade pizza-box
 *  systems. The CPLD has the  capability  to power
 *  down the system.  Arm and detonate.
 *
 * Input(s):
 *  None
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  Function does not return.
 *
 */
void
cpld_power_off(void)
{
#ifndef CONFIG_SILKWORM440_AP
	volatile char *cpld_base, *cpld_reset_enable, *cpld_reset_shutdown;

	cpld_base = (char *) ioremap(PULSAR_CPLD_BASE_ADDR, 32);
	if (cpld_base) {

		cpld_reset_enable = cpld_base + PULSAR_SYS_RESET_ENABLE_REG;
		cpld_reset_shutdown = cpld_base + PULSAR_SYS_RESET_SHUTDOWN_REG;

		/*
		 * Arm for system shutdown.
		 */
		writeb(PULSAR_SYSTEM_SHUTDOWN_ENABLE, cpld_reset_enable);
		
		/*
		 * Shutdown the system.
		 */
		writeb(PULSAR_SYSTEM_SHUTDOWN, cpld_reset_shutdown);

	} else {
		printk(KERN_CRIT "%s: ioremap failed\n", __FUNCTION__);
	}
#endif

}

void
cpld_power_off_odin(void)
{
#ifndef CONFIG_SILKWORM440_AP
        volatile char *cpld_base, *cpld_reset_enable, *cpld_reset_shutdown;

        cpld_base = (char *) ioremap(ODIN_CPLD_BASE_ADDR, 32);
        if (cpld_base) {

                cpld_reset_shutdown = cpld_base + ODIN_SYS_RESET_SHUTDOWN_REG;

                /*
                 * Shutdown the system.
                 */
                writeb(ODIN_SYSTEM_SHUTDOWN, cpld_reset_shutdown);

        } else {
                printk(KERN_CRIT "%s: ioremap failed\n", __FUNCTION__);
        }
#endif

}


EXPORT_SYMBOL(cpld_release_i2c_mux);
EXPORT_SYMBOL(cpld_arbitrate_i2c_mux);
EXPORT_SYMBOL(cpld_set_i2c_mux_local);
EXPORT_SYMBOL(cpld_set_i2c_mux);
EXPORT_SYMBOL(cpld_get_i2c_mux);
EXPORT_SYMBOL(cpld_spi_read);
EXPORT_SYMBOL(cpld_spi_write);
EXPORT_SYMBOL(silkworm);
EXPORT_SYMBOL(cpld_get_base_addr);
EXPORT_SYMBOL(cpld_check_pci_reset);
