/*
 * drivers/brcdecc/brcdecc.c
 *
 * Brocade's ECC reporting driver for Freescale 8458
 *
 * Author: Jeff Haran (jharan@brocade.com)
 *
 * Copyright (c) 2009 Brocade
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/mm.h>
#include <linux/device.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/dma-mapping.h>
#include <linux/crc32.h>
#include <linux/proc_fs.h>
#include <brocade/platform.h>
#include <brocade/silkworm.h>

/* in case DEBUG_PRINTKS gets defined in above includes */
#ifdef DEBUG_PRINTKS
#undef DEBUG_PRINTKS
#endif /* DEBUG_PRINTKS */

/*
 * remove DEBUG_PRINTKS define below to compile out debugging printks
 * in proc routines
 */
#define DEBUG_PRINTKS

/* module parameters */

static uint sbe_thresh = 1; /* what gets written to ERR_SBE[SBET] register */
static uint debug = 0;

/* definition of ccsr_ddr stolen from:
	/vobs/projects/springboard/prom/u_boot/include/asm-ppc/immap_85xx.h
	then changed uints to u32s and chars to u8s
 */
/* DDR memory controller registers(0x2000-0x3000) */

typedef struct ccsr_ddr {
	u32		cs0_bnds;		/* 0x2000 - DDR Chip Select 0 Memory Bounds */
	u8		res1[4];
	u32		cs1_bnds;		/* 0x2008 - DDR Chip Select 1 Memory Bounds */
	u8		res2[4];
	u32		cs2_bnds;		/* 0x2010 - DDR Chip Select 2 Memory Bounds */
	u8		res3[4];
	u32		cs3_bnds;		/* 0x2018 - DDR Chip Select 3 Memory Bounds */
	u8		res4[100];
	u32		cs0_config;		/* 0x2080 - DDR Chip Select Configuration */
	u32		cs1_config;		/* 0x2084 - DDR Chip Select Configuration */
	u32		cs2_config;		/* 0x2088 - DDR Chip Select Configuration */
	u32		cs3_config;		/* 0x208c - DDR Chip Select Configuration */
	u8		res5[112];
	u32		ext_refrec;		/* 0x2100 - DDR SDRAM Extended Refresh Recovery */
	u32		timing_cfg_0;		/* 0x2104 - DDR SDRAM Timing Configuration Register 0 */
	u32		timing_cfg_1;		/* 0x2108 - DDR SDRAM Timing Configuration Register 1 */
	u32		timing_cfg_2;		/* 0x210c - DDR SDRAM Timing Configuration Register 2 */
#define SDRAM_CFG_ECC_EN 0x20000000
	u32		sdram_cfg;		/* 0x2110 - DDR SDRAM Control Configuration */
	u32		sdram_cfg_2;		/* 0x2114 - DDR SDRAM Control Configuration 2 */
	u32		sdram_mode;		/* 0x2118 - DDR SDRAM Mode Configuration */
	u32		sdram_mode_2;		/* 0x211c - DDR SDRAM Mode Configuration 2*/
	u32		sdram_md_cntl;		/* 0x2120 - DDR SDRAM Mode Control */
	u32		sdram_interval;		/* 0x2124 - DDR SDRAM Interval Configuration */
	u32		sdram_data_init;	/* 0x2128 - DDR SDRAM Data initialization */
	u8		res6[4];
	u32		sdram_clk_cntl;		/* 0x2130 - DDR SDRAM Clock Control */
	u8		res7[20];
	u32		init_address;		/* 0x2148 - DDR training initialization address */
	u32		init_ext_address;	/* 0x214C - DDR training initialization extended address */
	u8		res8_1[2728];
	u32		ip_rev1;		/* 0x2BF8 - DDR IP Block Revision 1 */
	u32		ip_rev2;		/* 0x2BFC - DDR IP Block Revision 2 */
	u8		res8_2[512];
	u32		data_err_inject_hi;	/* 0x2e00 - DDR Memory Data Path Error Injection Mask High */
	u32		data_err_inject_lo;	/* 0x2e04 - DDR Memory Data Path Error Injection Mask Low */
	u32		ecc_err_inject;		/* 0x2e08 - DDR Memory Data Path Error Injection Mask ECC */
	u8		res9[20];
	u32		capture_data_hi;	/* 0x2e20 - DDR Memory Data Path Read Capture High */
	u32		capture_data_lo;	/* 0x2e24 - DDR Memory Data Path Read Capture Low */
	u32		capture_ecc;		/* 0x2e28 - DDR Memory Data Path Read Capture ECC */
	u8		res10[20];
#define ERR_DETECT_SBE 0x4
	u32		err_detect;		/* 0x2e40 - DDR Memory Error Detect */
	u32		err_disable;		/* 0x2e44 - DDR Memory Error Disable */
	u32		err_int_en;		/* 0x2e48 - DDR  */
#define CAP_ATTRIBS_VLD 0x1
	u32		capture_attributes;	/* 0x2e4c - DDR Memory Error Attributes Capture */
	u32		capture_address;	/* 0x2e50 - DDR Memory Error Address Capture */
	u32		capture_ext_address;	/* 0x2e54 - DDR Memory Error Extended Address Capture */
	u32		err_sbe;		/* 0x2e58 - DDR Memory Single-Bit ECC Error Management */
	u8		res11[164];
	u32		debug_1;		/* 0x2f00 */
	u32		debug_2;
	u32		debug_3;
	u32		debug_4;
	u8		res12[240];
} ccsr_ddr_t;

/*
 * not a lot of agreement in existing kernel sources as to what you should
 * return if a proc entry create fails. I found examples using -ENOMEM,
 * -ENODEV and -EIO. -ENOENT seems the most descriptive to me.
 */
#define PROC_CREATE_FAIL (-ENOENT)

/* pointer to virtual address of beginning of DDR controller */
static ccsr_ddr_t *p_ddr = NULL;

/* where we save the regs that we modify for restoration on exit */
static u32 err_disable_save;
static u32 err_int_en_save;
static u32 err_sbe_save;

/* did we write the ecc regs during init? */
static int ecc_regs_written = 0;

/* proc entry names and associated pointers */
#define ECC_PROC_DIR_NAME "driver/ecc"
static struct proc_dir_entry *ecc_dir = NULL;

#define ECC_PROC_MODEL_NAME "model"
static struct proc_dir_entry *proc_model = NULL;

#define ECC_PROC_ERR_DETECT_NAME "err_detect"
static struct proc_dir_entry *proc_err_detect = NULL;

#define ECC_PROC_ERR_DETECT_CLEAR_NAME "err_detect_clear"
static struct proc_dir_entry *proc_err_detect_clear = NULL;

#define ECC_PROC_ERR_DISABLE_NAME "err_disable"
static struct proc_dir_entry *proc_err_disable = NULL;

#define ECC_PROC_ERR_SBE_NAME "err_sbe"
static struct proc_dir_entry *proc_err_sbe = NULL;

#define ECC_PROC_ECC_ERR_INJECT_NAME "ecc_err_inject"
static struct proc_dir_entry *proc_ecc_err_inject = NULL;

/*
 * The read side of the procfs interface is screwy.
 * Read the comment block titled:
 * "How to be a proc read function" in fs/proc/generic.c:proc_file_read()
 * for some hint as to why the read functions below do what they do.
 *
 * Note that these implementations make some assumptions on the behavior
 * of the user space code reading them:
 *
 * 1) They read sequentially from the beginning of the file, no lseek()s.
 * 2) Their reads are big enough to get all the data in one read() call.
 *
 * These will be valid assumptions for the daemon written to read
 * these and cat, but perhaps not for other apps. User beware.
 */

/*
 * copy the lesser of d_len and s_len bytes from p_src to p_dst.
 * if s_len bytes are copied, indicate an end of file condition at *eof.
 *
 * return the number of bytes copied.
 */

static int proc_read_copy(
	char *p_dst, int d_len, char *p_src, int s_len, int *eof)
{
	int len = 0;

	while (d_len > 0) {
		if (s_len <= 0) {
			if (eof) {
				*eof = 1;
			}
			break;
		}
		*p_dst++ = *p_src++;
		d_len--;
		s_len--;
		len++;
	}
	return len;
}

#define MODEL_NAME "Freescale8548"

/* return the processor model to the caller */

static int proc_model_read(
	char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;

#ifdef DEBUG_PRINTKS
	if (debug > 1) {
		printk(KERN_DEBUG
			"proc_model_read(0x%p, 0x%p, 0x%lx, %d, 0x%x, 0x%p)\n",
				page, start ? *start : (char *) 0xdeadbeef, off,
				count, eof ? *eof : 0xdeadbeef, data);
	}
#endif /* DEBUG_PRINTKS */

	/* if this is not the first read, signal EOF to caller */
	if (*start || off) {
		*eof = 1;
		return 0;
	}

	len = proc_read_copy(page, count, MODEL_NAME, strlen(MODEL_NAME) + 1, eof);
	return len;
}

/* return the contents of the ERR_DETECT register to the caller */

static int proc_err_detect_read(
	char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;
	u32 err_detect;
	char buf[16];

#ifdef DEBUG_PRINTKS
	if (debug > 1) {
		printk(KERN_DEBUG
			"proc_err_detect_read(0x%p, 0x%p, 0x%lx, %d, 0x%x, 0x%p)\n",
				page, start ? *start : (char *) 0xdeadbeef, off,
				count, eof ? *eof : 0xdeadbeef, data);
	}
#endif /* DEBUG_PRINTKS */

	/* if this is not the first read, signal EOF to caller */
	if (*start || off) {
		*eof = 1;
		return 0;
	}

	err_detect = in_be32(&(p_ddr->err_detect));
	snprintf(buf, sizeof(buf), "%x", err_detect);
	len = proc_read_copy(page, count, buf, strlen(buf) + 1, eof);
	return len;
}

void edac_mc_scrub_block(unsigned long page, unsigned long offset, u32 size);

/*
 * the DDR controller will correct the data that it reads and presents on the bus,
 * but it wont write the corrected data back out the the DIMM. that's what this
 * function is for. borrows heavily from open source EDAC module.
 */

static void handle_sbe(void)
{
	u32 capture_attributes;
	u32 capture_address;
	unsigned long page_frame_number;
	unsigned long offset;

	capture_attributes = in_be32(&(p_ddr->capture_attributes));

#ifdef DEBUG_PRINTKS
	if (debug > 1) {
		printk(KERN_DEBUG
			"handle_sbe() attributes=0x%x\n", capture_attributes);
	}
#endif /* DEBUG_PRINTKS */

	/* the CAPTURE_ADDRESS isn't valid unless CAPTURE_ATTRIBUTES[VLD] is set */
	if (!(capture_attributes & CAP_ATTRIBS_VLD)) {
		return;
	}

	/* read the physical address of the data with the single bit error */
	capture_address = in_be32(&(p_ddr->capture_address));

	/* figure out what page frame number and offset into that page its at */
	page_frame_number = capture_address >> PAGE_SHIFT;
	offset = capture_address & (~PAGE_MASK);

#ifdef DEBUG_PRINTKS
	if (debug > 1) {
		printk(KERN_DEBUG
			"handle_sbe() address=0x%x, 0x%lx, 0x%lx\n",
				capture_address, page_frame_number, offset);
	}
#endif /* DEBUG_PRINTKS */

	/* read and write back the 8 bytes starting at that page and offset */
	edac_mc_scrub_block(page_frame_number, offset, 8);
	/* clear CAPTURE_ATTRIBUTES[VLD] */
	out_be32(&(p_ddr->capture_attributes), 0);
}

/*
 * return the contents of the ERR_DETECT register to the caller
 * and clear the contents of the register
 */

static int proc_err_detect_clear_read(
	char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;
	u32 err_detect;
	char buf[16];

#ifdef DEBUG_PRINTKS
	if (debug > 1) {
		printk(KERN_DEBUG
			"proc_err_detect_clear_read(0x%p, 0x%p, 0x%lx, %d, 0x%x, 0x%p)\n",
				page, start ? *start : (char *) 0xdeadbeef, off,
				count, eof ? *eof : 0xdeadbeef, data);
	}
#endif /* DEBUG_PRINTKS */

	/* if this is not the first read, signal EOF to caller */
	if (*start || off) {
		*eof = 1;
		return 0;
	}

	err_detect = in_be32(&(p_ddr->err_detect));
	if (err_detect) {
		if (err_detect & ERR_DETECT_SBE) {
			handle_sbe();
		}
		/* writing 1s bits clear them */
		out_be32(&(p_ddr->err_detect), err_detect);
	}
	snprintf(buf, sizeof(buf), "%x", err_detect);
	len = proc_read_copy(page, count, buf, strlen(buf) + 1, eof);
	return len;
}

/* return the contents of the ERR_DISABLE register to the caller */

static int proc_err_disable_read(
	char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;
	u32 err_disable;
	char buf[16];

#ifdef DEBUG_PRINTKS
	if (debug > 1) {
		printk(KERN_DEBUG
			"proc_err_disable_read(0x%p, 0x%p, 0x%lx, %d, 0x%x, 0x%p)\n",
				page, start ? *start : (char *) 0xdeadbeef, off,
				count, eof ? *eof : 0xdeadbeef, data);
	}
#endif /* DEBUG_PRINTKS */

	/* if this is not the first read, signal EOF to caller */
	if (*start || off) {
		*eof = 1;
		return 0;
	}

	err_disable = in_be32(&(p_ddr->err_disable));
	snprintf(buf, sizeof(buf), "%x", err_disable);
	len = proc_read_copy(page, count, buf, strlen(buf) + 1, eof);
	return len;
}

/*
 * scan up to the first 8 chars of buf into their hex representation
 * and return the result.
 */

static u32 scan_u32(char *buf)
{
	u32 val = 0;
	int len = 0;
	char *p = buf;

	for ( ; len < 8; ++p, ++len ) {
		if ((*p >= '0') && (*p <= '9')) {
			val = (val << 4) + *p - '0';
			continue;
		}
		if ((*p >= 'a') && (*p <= 'f')) {
			val = (val << 4) + *p + 10 - 'a';
			continue;
		}
		if ((*p >= 'A') && (*p <= 'F')) {
			val = (val << 4) + *p + 10 - 'A';
			continue;
		}
		break;
	}
	return val;
}

/*
 * convert the user written string of hex characters to binary
 * and write to the ERR_DISABLE register.
 */

static int proc_err_disable_write(
	struct file *file, const char *buffer, unsigned long count, void *data)
{
	int len;
	u32 err_disable;
	char my_buf[16];

	if (count > (sizeof(my_buf) - 1)) {
		len = sizeof(my_buf) - 1;
	} else {
		len = count;
	}

	if (copy_from_user(my_buf, buffer, len)) {
		return -EFAULT;
	}

	my_buf[len] = 0;

	err_disable = scan_u32(my_buf);

	out_be32(&(p_ddr->err_disable), err_disable);
	return len;
}

/* return the contents of the ERR_SBE register to the caller */

static int proc_err_sbe_read(
	char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;
	u32 err_sbe;
	char buf[16];

#ifdef DEBUG_PRINTKS
	if (debug > 1) {
		printk(KERN_DEBUG
			"proc_err_sbe_read(0x%p, 0x%p, 0x%lx, %d, 0x%x, 0x%p)\n",
				page, start ? *start : (char *) 0xdeadbeef, off,
				count, eof ? *eof : 0xdeadbeef, data);
	}
#endif /* DEBUG_PRINTKS */

	/* if this is not the first read, signal EOF to caller */
	if (*start || off) {
		*eof = 1;
		return 0;
	}

	err_sbe = in_be32(&(p_ddr->err_sbe));
	snprintf(buf, sizeof(buf), "%x", err_sbe);
	len = proc_read_copy(page, count, buf, strlen(buf) + 1, eof);
	return len;
}

/*
 * convert the user written string of hex characters to binary
 * and write to the ERR_SBE register.
 */

static int proc_err_sbe_write(
	struct file *file, const char *buffer, unsigned long count, void *data)
{
	int len;
	u32 err_sbe;
	char my_buf[16];

	if (count > (sizeof(my_buf) - 1)) {
		len = sizeof(my_buf) - 1;
	} else {
		len = count;
	}

	if (copy_from_user(my_buf, buffer, len)) {
		return -EFAULT;
	}

	my_buf[len] = 0;

	err_sbe = scan_u32(my_buf);

	out_be32(&(p_ddr->err_sbe), err_sbe);
	return len;
}

/* return the contents of the ECC_ERR_INJECT register to the caller */

static int proc_ecc_err_inject_read(
	char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;
	u32 ecc_err_inject;
	char buf[16];

#ifdef DEBUG_PRINTKS
	if (debug > 1) {
		printk(KERN_DEBUG
			"proc_ecc_err_inject_read(0x%p, 0x%p, 0x%lx, 0x%x, 0x%x, 0x%p)\n",
				page, start ? *start : (char *) 0xdeadbeef, off,
				count, eof ? *eof : 0xdeadbeef, data);
	}
#endif /* DEBUG_PRINTKS */

	/* if this is not the first read, signal EOF to caller */
	if (*start || off) {
		*eof = 1;
		return 0;
	}

	ecc_err_inject = in_be32(&(p_ddr->ecc_err_inject));
	snprintf(buf, sizeof(buf), "%x", ecc_err_inject);
	len = proc_read_copy(page, count, buf, strlen(buf) + 1, eof);
	return len;
}

/*
 * convert the user written string of hex characters to binary
 * and write to the ECC_ERR_INJECT register.
 */

static int proc_ecc_err_inject_write(
	struct file *file, const char *buffer, unsigned long count, void *data)
{
	int len;
	u32 ecc_err_inject;
	char my_buf[16];

	if (count > (sizeof(my_buf) - 1)) {
		len = sizeof(my_buf) - 1;
	} else {
		len = count;
	}

	if (copy_from_user(my_buf, buffer, len)) {
		return -EFAULT;
	}

	my_buf[len] = 0;

	ecc_err_inject = scan_u32(my_buf);

	out_be32(&(p_ddr->ecc_err_inject), ecc_err_inject);
	return len;
}

/* deallocate/destroy stuff created in init routine */

static void brcd_ecc_cleanup(void)
{
	if (ecc_regs_written) {
		/* restore insmod time register values */
		out_be32(&(p_ddr->err_int_en), err_int_en_save);
		out_be32(&(p_ddr->err_disable), err_disable_save);
		out_be32(&(p_ddr->err_sbe), err_sbe_save);

		ecc_regs_written = 0;
	}

	if (proc_ecc_err_inject) {
		remove_proc_entry(ECC_PROC_ECC_ERR_INJECT_NAME, ecc_dir);
		proc_ecc_err_inject = NULL;
	}

	if (proc_err_sbe) {
		remove_proc_entry(ECC_PROC_ERR_SBE_NAME, ecc_dir);
		proc_err_sbe = NULL;
	}

	if (proc_err_disable) {
		remove_proc_entry(ECC_PROC_ERR_DISABLE_NAME, ecc_dir);
		proc_err_disable = NULL;
	}

	if (proc_err_detect_clear) {
		remove_proc_entry(ECC_PROC_ERR_DETECT_CLEAR_NAME, ecc_dir);
		proc_err_detect_clear = NULL;
	}

	if (proc_err_detect) {
		remove_proc_entry(ECC_PROC_ERR_DETECT_NAME, ecc_dir);
		proc_err_detect = NULL;
	}

	if (proc_model) {
		remove_proc_entry(ECC_PROC_MODEL_NAME, ecc_dir);
		proc_model = NULL;
	}

	if (ecc_dir) {
		remove_proc_entry(ECC_PROC_DIR_NAME, NULL);
		ecc_dir = NULL;
	}

	if (p_ddr) {
		iounmap(p_ddr);
		p_ddr = NULL;
	}
}

static int __init brcdecc_init(void)
{
	int rc = 0;

	printk(KERN_INFO "brcdecc_init() %u %u 0x%x\n",
			debug, sbe_thresh, BOARD_CCSRBAR);

	/* only low 8 bits of sbe_thresh are significant */
	if (sbe_thresh & 0xffffff00) {
		sbe_thresh = 0xff;
		printk(KERN_WARNING "sbe_thresh truncated to %u\n", sbe_thresh);
	}

	/* we break out of the following one pass "do" on errors */
	do {
		/* get a pointer to the DDR register bank */
		p_ddr = (ccsr_ddr_t *) ioremap_nocache(BOARD_CCSRBAR + 0x2000, 0x1000);
		if (!p_ddr) {
			printk(KERN_ERR "brcdecc_init() ioremap_nocache() failed\n");
			rc = -ENOMEM;
			break;
		}

		if ((((char *) &(p_ddr->err_sbe)) - ((char *) p_ddr)) != 0xe58) {
			/* this would mean the definition of ccsr_ddr is wrong */
			printk(KERN_ERR "fix struct ccsr_ddr 0x%x\n",
				(((char *) &(p_ddr->err_sbe)) - ((char *) p_ddr)));
			rc = -ENODEV;
			break;
		}

		/*
		 * make sure the PROM enabled ECC. we can't since this bit has
		 * be set before code starts executing from DRAM.
		 */
		if (!(in_be32(&(p_ddr->sdram_cfg)) & SDRAM_CFG_ECC_EN)) {
			printk(KERN_ERR "brcdecc_init() ECC_EN not enabled\n");
			rc = -ENODEV;
			break;
		}

		/* create /proc/driver/ecc directory */
		ecc_dir = proc_mkdir(ECC_PROC_DIR_NAME, NULL);
		if (!ecc_dir) {
			printk(KERN_ERR "Failed to create ecc proc dir\n");
			rc = PROC_CREATE_FAIL;
			break;
		}
		ecc_dir->owner = THIS_MODULE;

		/* create ecc/model, readonly to everybody */
		proc_model = create_proc_entry(ECC_PROC_MODEL_NAME, 0444, ecc_dir);
		if (!proc_model) {
			printk(KERN_ERR "Failed to create ecc proc model\n");
			rc = PROC_CREATE_FAIL;
			break;
		}
		proc_model->data = p_ddr;
		proc_model->owner = THIS_MODULE;
		proc_model->read_proc = proc_model_read;
		proc_model->write_proc = NULL;

		/* create ecc/err_detect, readonly to everybody */
		proc_err_detect =
			create_proc_entry(ECC_PROC_ERR_DETECT_NAME, 0444, ecc_dir);
		if (!proc_err_detect) {
			printk(KERN_ERR "Failed to create ecc proc err_detect\n");
			rc = PROC_CREATE_FAIL;
			break;
		}
		proc_err_detect->data = p_ddr;
		proc_err_detect->owner = THIS_MODULE;
		proc_err_detect->read_proc = proc_err_detect_read;
		proc_err_detect->write_proc = NULL;

		/* create ecc/err_detect_clear, readonly to root only */
		proc_err_detect_clear =
			create_proc_entry(ECC_PROC_ERR_DETECT_CLEAR_NAME, 0400, ecc_dir);
		if (!proc_err_detect_clear) {
			printk(KERN_ERR "Failed to create ecc proc err_detect_clear\n");
			rc = PROC_CREATE_FAIL;
			break;
		}
		proc_err_detect_clear->data = p_ddr;
		proc_err_detect_clear->owner = THIS_MODULE;
		proc_err_detect_clear->read_proc = proc_err_detect_clear_read;
		proc_err_detect_clear->write_proc = NULL;

		/* create ecc/err_disabled, readonly to everybody, unless debug on */
		proc_err_disable =
			create_proc_entry(ECC_PROC_ERR_DISABLE_NAME,
				(debug ? 0644 : 0444), ecc_dir);
		if (!proc_err_disable) {
			printk(KERN_ERR "Failed to create ecc proc err_disable\n");
			rc = PROC_CREATE_FAIL;
			break;
		}
		proc_err_disable->data = p_ddr;
		proc_err_disable->owner = THIS_MODULE;
		proc_err_disable->read_proc = proc_err_disable_read;
		if (debug) {
			proc_err_disable->write_proc = proc_err_disable_write;
		} else {
			proc_err_disable->write_proc = NULL;
		}

		/* create ecc/err_sbe, readonly to everybody, unless debug on */
		proc_err_sbe =
			create_proc_entry(ECC_PROC_ERR_SBE_NAME,
				(debug ? 0644 : 0444), ecc_dir);
		if (!proc_err_sbe) {
			printk(KERN_ERR "Failed to create ecc proc err_sbe\n");
			rc = PROC_CREATE_FAIL;
			break;
		}
		proc_err_sbe->data = p_ddr;
		proc_err_sbe->owner = THIS_MODULE;
		proc_err_sbe->read_proc = proc_err_sbe_read;
		if (debug) {
			proc_err_sbe->write_proc = proc_err_sbe_write;
		} else {
			proc_err_sbe->write_proc = NULL;
		}

		if (debug) {
			/* create ecc/ecc_err_inject, read-write to root */
			proc_ecc_err_inject =
				create_proc_entry(ECC_PROC_ECC_ERR_INJECT_NAME, 0644, ecc_dir);
			if (!proc_ecc_err_inject) {
				printk(KERN_ERR "Failed to create ecc proc ecc_err_inject\n");
				rc = PROC_CREATE_FAIL;
				break;
			}
			proc_ecc_err_inject->data = p_ddr;
			proc_ecc_err_inject->owner = THIS_MODULE;
			proc_ecc_err_inject->read_proc = proc_ecc_err_inject_read;
			proc_ecc_err_inject->write_proc = proc_ecc_err_inject_write;
		}

		/* save values in these regs for restoration on module exit */
		err_disable_save = in_be32(&(p_ddr->err_disable));
		err_int_en_save = in_be32(&(p_ddr->err_int_en));
		err_sbe_save = in_be32(&(p_ddr->err_sbe));

		/* no interrupts on errors */
		out_be32(&(p_ddr->err_int_en), 0);
		/* enable reporting of all errors */
		out_be32(&(p_ddr->err_disable), 0);
		/* set threshold to user specified value */
		out_be32(&(p_ddr->err_sbe), ((sbe_thresh & 0xff) << 16));

		/* flag that we got far enough to write the ECC regs */
		ecc_regs_written = 1;

	} while (0);

	/* on any error, cleanup the mess */
	if (rc) {
		brcd_ecc_cleanup();
	}

	return rc;
}

static void __exit brcdecc_exit(void)
{
	brcd_ecc_cleanup();
}

module_init(brcdecc_init);
module_exit(brcdecc_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Brocade");
module_param(sbe_thresh, uint, 0644);
MODULE_PARM_DESC(sbe_thresh, "Value written to ERR_SBE register SBET field");
module_param(debug, uint, 0644);
MODULE_PARM_DESC(debug, "Set to non-zero to enable some debugging stuff");

