/*
 *
 *    Copyright (c) 1996-2008 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *    Description:
 *      Silkworm Platform device driver.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/miscdevice.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/rtc.h>
#include <linux/i2c.h>

#include <asm/time.h>
#include <brocade/cpld.h>
#include <brocade/platform.h>
#include <brocade/silkworm.h>
#include <platforms/4xx/silkworm_sin.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>


#define PLATFORM_VERSION		"1.0"

extern sw_bd silkworm;

static const char *platform_name = "platform";
static int platform_major = 0;

/* gpb_ = generic brocade platform */
static unsigned int
gbp_get_swbd(void) {
	return (silkworm.board);
}

static int
gbp_get_cpid(unsigned long *result) {
	int status = 0;
	unsigned char reg_data;

	switch (silkworm.board) {
		
	case SWBD63:
		/* Zentron Co-CPU */
		status = cpld_read_register(MET_BOOTROM_CTRL, &reg_data);
		if (0 == status) {
			/* the mask is one bit */
			*result = (reg_data & COCPU_CP_SLOT_ID) ? 1 : 0;
		}
		break;

	case SWBD68:
	case SWBD67:
	case SWBD62:
	case SWBD42:
	case SWBD46:
	case SWBD64:
    case SWBD76:
  	case SWBD77:
	case SWBD133:
		/* Saturn registers are same as Meteor */
		/* Fall Through */
	case SWBD21:
		status = cpld_read_register(MET_CP_MISC_READ, &reg_data);
		if (0 == status) {
			/* the mask is one bit */
			*result = (reg_data & MET_CP_CP_SLOT_ID) ? 1 : 0;
		}
		break;
	case SWBD10:
		status = cpld_read_register(MISC_READ_OFFSET, &reg_data);
		if (0 == status) {
			/* the mask is one bit */
			*result = (reg_data & CPLD_SLOT_ID_FIELD) ? 1 : 0;
		}
		break;
	default:
		*result = 0;
		status = 0;
		break;
	}

	return (status);
}

/*
 * int platform_ioctl()
 *
 * Description:
 *   Implements the IOCTL interface to obtain platform
 *   specific information
 *
 *   IOCTLs supported are:
 *     SWBD_IOC_GET_HWMODEL  - Get the Hardware model ID number
 *			       as defined in include/brocade/silkworm.h
 *
 *     SWBD_IOC_GET_REVISION - Get the hardware board revision.
 *			       Valid values are 0-7
 *
 *     SWBD_IOC_GET_CPID     - Get the CP ID or slot number.
 *			       Valid values are 0 for CP0 (slot 5)
 *			       or 1 for CP1 (slot 6). For single
 *			       board systems, the value is 0.
 *
 * Input(s):
 *   inode  - Pointer to the inode descriptor for the device
 *   file   - Pointer to an open file descriptor for the device
 *   cmd    - The IOCTL command
 *   arg    - User-space data
 *
 * Output(s):
 *   See above
 *
 * Returns:
 *    0 - Success
 *   -EFAULT - Fail
 *
 */

static int
platform_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		unsigned long arg)
{
	int status;
	unsigned long data;

	switch (cmd) {


	case SWBD_IOC_GET_HWMODEL:

		data = (long)gbp_get_swbd();	/* silkworm.board */
		break;


	case SWBD_IOC_GET_REVISION:

		data = (long)silkworm.board_rev;
		break;

	case SWBD_IOC_GET_CPID:
	{
#ifdef	CONFIG_SILKWORM440_AP
		data = 0;
		status = 0;
#else
		status = gbp_get_cpid(&data);

		if (status)	/* preserve ancestors behavior */
			return (-EFAULT);

#endif /* !CONFIG_SILKWORM440AP */
		break;
	}
	case SWBD_IOC_GET_SINSTATS:
		return ioctl_sinstats((void **)&arg);

	case SWBD_IOC_GET_SINTBL:
		return ioctl_sintbl((void **)&arg);

	default:
		return (-EINVAL);
	}

	status = copy_to_user((void *)arg, &data, sizeof (data));

	return (status ? -EFAULT : 0);
}


static struct file_operations platform_fops = {
	owner:	THIS_MODULE,
	ioctl:	platform_ioctl,

};


/*
 * int __init platform_init()
 *
 * Description:
 *   Initialzation function for the m41t11 device driver
 *
 * Input(s):
 *   N/A
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *   0 = Success
 *   NonZero = Fail
 *
 */
static int __init
platform_init(void)
{
	if ((platform_major = register_chrdev(platform_major, platform_name,
					&platform_fops)) < 0) {

		printk("platform_init: cannot register device (%d) %s",
				current->pid, platform_name);
		return (-ENOMEM);
	}

	/*
	 * We have to support CP03 for a while, and they
	 * don't pass platform info to the kernel. So
	 * if the board model is a 0, we assume a
	 * Ulysses with a CP03. The board rev is 1.
	 *
	 */

	if (silkworm.board == 0) {

		silkworm.board = SWBD10;
		silkworm.board_rev = 1;

	}
	printk(KERN_INFO "SWBD Platform "
		"Driver v" PLATFORM_VERSION ": [type %d, rev %d].\n", 
		silkworm.board, silkworm.board_rev);

	printk(KERN_INFO "Config Silkworm \n");

	return (0);
}

/*
 * void __exit platform_exit()
 *
 * Description
 *   Function to remove the device driver. Removes
 *   the proc file system entry and un-registers
 *   the file ops.
 *
 * Input(s):
 *   N/A
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *   N/A
 *
 */
static void __exit
platform_exit(void)
{

	unregister_chrdev(platform_major, platform_name);
}

EXPORT_SYMBOL(gbp_get_swbd);
EXPORT_SYMBOL(gbp_get_cpid);

module_init(platform_init);
module_exit(platform_exit);
