/*
 *
 *    Copyright (c) 2008 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *    Description:
 *      Intel/DEC 21554 non-transparent PCI-to-PCI bridge (aka Drawbridge)
 *      device driver for 440gp-based directors.
 *
 */
#include <linux/types.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/version.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>
#include <asm/processor.h>


#include <brocade/cpld.h>
#include <brocade/swbd.h>

#include <platforms/4xx/silkworm_sin.h>	/* for sin_get_swbd() */

/*
 * Please include the righ header for the right platform.
 */

#include <asm/ibm4xx.h>
#include <platforms/4xx/ibm440gp.h>	/* for GPIO base addr */

#include "../include/db.h"


/* Preprocessor Defines */


/* Function Macros */

#ifdef DEBUG
#define DBGPRINT(x...) printk(KERN_ERR x)
#else
#define DBGPRINT(x...)
#endif

extern db_known_dev_t common_known_devs[];
/* Pointer to current platform configuration instance.  */
extern db_platform_cfg_t *cur_cfg_ptr;

/* Bit masks for the GPIO pins that control primary side DB reset */
#define GPIO0_BIT_18		0x00002000
#define GPIO0_BIT_19		0x00001000
#define GPIO0_BIT_20		0x00000800

/* Mask for the GPIO Trace Enable field */
#define CPC0_GPIO_TRE_MASK  	0x00002000	/* 440GP CPC0_GPIO */

/* PPC440 GPIO control register offsets */
#define GPIO0_REG_MAP_SZ	0x10	/* Size of register space to map */
#define GPIO0_OR_OFFSET		0x00	/* Output Register */
#define GPIO0_TCR_OFFSET	0x04	/* Three-State Control Register */




static int db_primary_reset_440(u_long gpio_mask);
static int db_meteor_primary_reset(void);

static const uint32_t meteor_base_addrs[] = {
	0x40000000, 				/* drawbridge 1 */
	0x70000000,  				/* drawbridge 2 */
	0xA0000000   				/* drawbridge 3 */
};

static db_platform_cfg_t meteor_cfg = {		/* METEOR */
    swbd: SWBD_METEOR,
    num_db: 3,					/* num drawbridges */
    busses_per_slot: 10,			/* busses per chassis slot */
    root_bus_incr: 0x40,			/* root bus increment */
    irq_num: 26,				/* internal IRQ num */
    cp_subordinate_bus: 1,			/* Default doesn't matter */
    physical_pci_base: METEOR_PCI_PHYS_BASE,
    physical_db_base: DB_METEOR_BASE,
    device_alloc_sz: 0,				/* N/A */
    blade_alloc_sz: DB_METEOR_BLADE_ALLOTMENT, /* Allow 128MB per blade slot */
    fwd_win_offset: 0x08000000,			/* forwarding window offset */
	gpio_base: NULL,				/* to be initialized */
    base_addr: meteor_base_addrs, 		/* drawbridge base addresses */
    db_fixup_method: db_fixup_drawbridges,		/* Fix-up routine */
    config_bases_method: db_configure_bases,	/* Always configure BARs */
    read_bases_method: db_read_bases,		/* Read cur config of BARs */

    setup_dma_method: NULL,			/* No ASIC DMA setup */
    reset_method: db_meteor_primary_reset,	/* ptr to reset func */
    alloc_method: db_dynamic_address_allocate, 	/* ptr to addr. alloc method */
	ntb_lt_primary: 0x10,			/* allow up to n data transactions */
	ntb_lt_secondary: 0x10,			/* allow up to n data transactions */
	tb_lt_primary: 0x10,			/* allow up to n data transactions */
	tb_lt_secondary: 0x10,			/* allow up to n data transactions */
	enable_secondary_mem: 0,		/* disable DB secondary side mem */
	known_devs: NULL,				/* list of devices known apriori */

};

/* Internal configuration table. */
db_platform_cfg_t *(platform_cfg_tbl[]) = {

	&meteor_cfg,

};

int db_cfg_tbl_sz = (sizeof (platform_cfg_tbl) / sizeof (void *));

/*
 *
 * int db_primary_reset_init()
 *
 * Description:
 *
 *  Encapsulation of the processor-specific initialization in
 *  support of the DB primary bus interface reset.  The reset
 *  control function should only *control* the interface, not
 *  initialize it: including mapping/unmapping resources.
 *
 *  A subsequent designer/developer may wish to generalize this
 *  to support all platform/processor-specific initialization
 *
 * Input(s):
 *  None.
 *
 * Output(s):
 *  global gpio_base member of DB control structure is initialized
 *  with virtual memory base address
 *
 * Returns:
 *  0 - Success
 *  -ENODEV - no device control structure found
 *  -ENOMEM - i/o subsystem could not map resources
 *
 */
int
db_primary_reset_init(void)
{
  /* Map the GPIO control register space. */
	if (cur_cfg_ptr == NULL) {
		return (-ENODEV);
	}

	cur_cfg_ptr->gpio_base =
		ioremap_nocache(PPC440GP_GPIO0_ADDR, GPIO0_REG_MAP_SZ);

	return ((cur_cfg_ptr->gpio_base != NULL) ? 0 : -ENOMEM);
}

/*
 *
 * int db_primary_reset_440()
 *
 * Description:
 *
 *  Reset the specified 21555 drawbridges by toggling their corresponding
 *  GPIO lines that are tied to the 21555 primary reset input.
 *  A 0 asserts reset, a 1 de-asserts reset. Asserting this reset clears
 *  all configuration registers, so the registers must be saved and then
 *  restored.
 *
 * To avoid confusion over which bit to toggle and when,
 * we reset and restore all drawbridges.
 *
 * NOTE: The configuration registers of both bridges should
 *       be saved prior to calling this routine.
 *
 *       This routine is CPU-specific because it uses CPU
 *       GPIO registers.
 *
 *       This routine supports both the 440GP and 440GX processors.
 *
 * Input(s):
 *  Bit mask indicating which GPIO lines are to be toggled.
 *
 * Output(s):
 *  N/A
 *
 * Returns:
 *  0 - Success
 *  Non-0 - Failed.
 *
 */
static int
db_primary_reset_440(u_long gpio_mask)
{
  u_long gpio_or, gpio_tcr, flags; /* DO NOT DELETE. USED BY SYS_INT_SAVE() */
  int error = 0;
  void *gpio_base;


  DBGPRINT("PCI DrawBridge Primary Reset GPIO Word 0x%lx\n", gpio_mask);

  if ((cur_cfg_ptr == NULL) || (cur_cfg_ptr->gpio_base == NULL)) {
	return (-ENODEV);
  }
  gpio_base = (void *)cur_cfg_ptr->gpio_base;

  gpio_or = in_be32((volatile __u32 *)(gpio_base + GPIO0_OR_OFFSET));
  gpio_tcr = in_be32((volatile __u32 *)(gpio_base + GPIO0_TCR_OFFSET));

  DBGPRINT("Current Output Reg 0x%8.8lx  TCR: 0x%8.8lx\n", gpio_or, gpio_tcr);

  DBGPRINT("Proposed reset ON  word: 0x%8.8lx\n", gpio_or & ~gpio_mask);
  DBGPRINT("Proposed TCS word:       0x%8.8lx\n", gpio_tcr | gpio_mask);
  DBGPRINT("Proposed reset OFF word: 0x%8.8lx\n", gpio_or | gpio_mask);


  /* Force a bridge reset via the 440 gpio0 register. Clears */
  /* all config registers, and re-reads strapping pins. */
  /* RMW GPIO output register to force the appropriate bits */
  /* low, turn off tristate, then force the bits back high */

  SYS_INT_SAVE();

  out_be32((volatile __u32 *)(gpio_base + GPIO0_OR_OFFSET),
	  (gpio_or & ~gpio_mask));
  out_be32((volatile __u32 *)(gpio_base + GPIO0_TCR_OFFSET),
	  gpio_tcr | gpio_mask);

  /*
   * Don't allow anything else to run. The Drawbridges
   * are unusable at this point and we could hang if
   * anyone tries a PCI access.
   * But don't let the delay to exceed 1ms.
   */
  udelay(1000);

  out_be32((volatile __u32 *)(gpio_base + GPIO0_OR_OFFSET),
	  (gpio_or | gpio_mask));

  SYS_INT_RESTORE();

  DBGPRINT("Final GPIO0_OR 0x%8.8x  GPIO0_TCR 0x%8.8x\n",
      in_be32((volatile __u32 *)(gpio_base + GPIO0_OR_OFFSET)),
      in_be32((volatile __u32 *)(gpio_base + GPIO0_TCR_OFFSET)));

  return (error);

}


/*
 *
 * int db_meteor_primary_reset()
 *
 * Description:
 *
 *  Bits 18, 19, and 20 of the 440 GPIO register are used to toggle
 *  the 21555 primary reset lines on Meteor.
 *
 *  The GPIO bits are connected as:
 *    Bit 18 resets 21555 device 1
 *    Bit 19 resets 21555 device 2
 *    Bit 20 resets 21555 device 3
 *
 * NOTE: The configuration registers of all bridges should
 *       be saved prior to calling this routine.
 *
 *       This routine is CPU-specific because it uses CPU
 *       GPIO registers.
 *
 * Input(s):
 *  N/A
 *
 * Output(s):
 *  N/A
 *
 * Returns:
 *  0 - Success
 *  Non-0 - Failed.
 *
 */
int
db_meteor_primary_reset(void)
{

  u_long flags;		/* DO NOT DELETE. USED BY SYS_INT_SAVE() */
  uint32_t temp;

  /* The reset lines have dual usage--trace or GPIO.  Configure them as GPIO. */
  SYS_INT_SAVE();
  temp = mfdcr(DCRN_CPC0_GPIO);
  temp &= ~CPC0_GPIO_TRE_MASK;
  mtdcr(DCRN_CPC0_GPIO, temp);
  SYS_INT_RESTORE();


  return (db_primary_reset_440(GPIO0_BIT_18 | GPIO0_BIT_19 | GPIO0_BIT_20));
}





