/*
 *  Copyright  2001 Flaga hf. Medical Devices, Kri Davsson <kd@flaga.is>
 *
 *  $Id: cfi_silkworm.c,v 1.7 2001/10/02 15:05:13 dwmw2 Exp $
 *  
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <brocade/platform.h>
#include <brocade/swbd.h>

extern sw_bd silkworm;

static uint8_t *cpld_addr;
static uint8_t cfi_vpp_offset, cfi_vpp_mask;
static uint32_t cfi_vpp_offset32, cfi_vpp_mask32;
static uint32_t jedec_vpp_offset32, jedec_vpp_mask32, jedec_vpp_magic32;
static uint8_t jedec_vpp_offset, jedec_vpp_mask;
static struct my_mtd_mapping_t *silkworm_maps;


/* This is a local struct used to help simplify the assignment of MTD maps/partitions.
 */
struct my_mtd_mapping_t {
	struct map_info		*map;		/* pointer to our chip */
	struct mtd_partition	*parts;		/* pointer to array of paritions */
	struct mtd_info		*mtd_handle;	/* this is assigned the result of do_probe() */
	int			nparts;		/* number of partitions in part array above */
	char			*probe;		/* string name to use for probe */
	void			*base;		/* base address of flash chip */
};

#ifndef elemof
#define elemof(a) (sizeof(a)/sizeof((a)[0]))
#endif


/* These are the prefixes to the MTD name field, which is used by ramdisk scripts to generate /dev symlinks
 * at runtime.  See the document Linux Infrastructure Support for Tsunami for more details.
 */
#define KERN0_STR			"kern0: "
#define KERN1_STR			"kern1: "
#define PROM_STR			"prom: "	/* sys w/ only one prom */
#define PROM0_STR			"prom0: "	/* sys w/ two proms */
#define PROM1_STR			"prom1: "	/* sys w/ two proms */
#define BOOTENV_STR			"bootenv: "
#define BOOTENV0_STR			"bootenv0: "
#define BOOTENV1_STR			"bootenv1: "
#define BOOTSEL_STR			"bootsel: "
#define DUMP0_STR			"dump0: "
#define DUMP1_STR			"dump1: "
#define PSEUDO_BOOTENV0_STR		"bootenv0s: "
#define PSEUDO_BOOTENV1_STR		"bootenv1s: "


/* -------------------------------------------------------------------------
 * ulysses
 * Flash information for ulysses aka silkworm 12000.
 * 512 KB of boot flash of type JEDEC.
 * 64 MB of user flash of type CFI.
 *------------------------------------------------------------------------*/

/* JEDEC boot PROM */
# define ULYSSES_MTD_JEDEC_START		0xFFF80000
# define ULYSSES_MTD_JEDEC_LEN			0x00080000
# define ULYSSES_MTD_JEDEC_BUSWIDTH		(1)

/* CFI bulk flash */
# define ULYSSES_MTD_CFI_START			0xF0000000
# define ULYSSES_MTD_CFI_LEN			0x01000000
# define ULYSSES_MTD_CFI_BUSWIDTH		(4)

/* CPLD information. These are register which control write access
   to the flash chips. */
#define ULYSSES_CPLD_BASE_ADDRESS		0xF8000000
#define ULYSSES_CPLD_SIZE			0x20
#define ULYSSES_CPLD_CFI_WRITE_ENABLE_OFFSET	0x1C
#define ULYSSES_CPLD_CFI_WRITE_ENABLE_BITS	0x03
#define ULYSSES_CPLD_JEDEC_WRITE_ENABLE_OFFSET	0x1B
#define ULYSSES_CPLD_JEDEC_WRITE_ENABLE_BITS	0x01

/* -------------------------------------------------------------------------
 * warp
 * Flash information for warp.
 * 512 KB of boot flash of type JEDEC.
 * 8 MB of user flash of type CFI.
 *------------------------------------------------------------------------*/

/* JEDEC boot PROM */
# define WARP_MTD_JEDEC_START		0xFFF80000
# define WARP_MTD_JEDEC_LEN	       	0x00080000
# define WARP_MTD_JEDEC_BUSWIDTH       	(1)

/* CFI bulk flash */
# define WARP_MTD_CFI_START	       	0x40000000
# define WARP_MTD_CFI_LEN	       	0x00800000
# define WARP_MTD_CFI_BUSWIDTH		(4)

/* -------------------------------------------------------------------------
 * terminator
 * Flash information for terminator.
 * 512 KB of boot flash of type JEDEC.
 * 64 MB of user flash of type CFI.
 *------------------------------------------------------------------------*/

/* JEDEC boot PROM */
# define TERMINATOR_MTD_JEDEC_START		0xFFF80000
# define TERMINATOR_MTD_JEDEC_LEN	       	0x00080000
# define TERMINATOR_MTD_JEDEC_BUSWIDTH		(1)

/* CFI bulk flash */
# define TERMINATOR_MTD_CFI_START		0xF0000000
# define TERMINATOR_MTD_CFI_LEN			0x01000000
# define TERMINATOR_MTD_CFI_BUSWIDTH		(4)

/* CPLD information. These are register which control write access
   to the flash chips. */
#define TERMINATOR_CPLD_BASE_ADDRESS		0xF8000000
#define TERMINATOR_CPLD_SIZE			0x100
#define TERMINATOR_CPLD_CFI_WRITE_ENABLE_OFFSET	0x94
#define TERMINATOR_CPLD_CFI_WRITE_ENABLE_BITS	0x03
#define TERMINATOR_CPLD_JEDEC_WRITE_ENABLE_OFFSET 0x90
#define TERMINATOR_CPLD_JEDEC_WRITE_ENABLE_BITS	0x01

/* -------------------------------------------------------------------------
 * meteor
 * Flash information for meteor aka silkworm 12000.
 * 512 KB of boot flash of type JEDEC.
 * 64 MB of user flash of type CFI.
 *------------------------------------------------------------------------*/

/* JEDEC boot PROM */
# define METEOR_MTD_JEDEC_START			0x00000001FFF80000 
# define METEOR_MTD_JEDEC_LEN			0x00080000
# define METEOR_MTD_JEDEC_BUSWIDTH		(1)

/* CFI bulk flash chip 1 for Meteor */
# define METEOR_MTD_CFI_LOW_START		0x00000001F0000000 
# define METEOR_MTD_CFI_LOW_LEN			0x00800000
# define METEOR_MTD_CFI_LOW_BUSWIDTH		(2)

/* CFI bulk flash chip 2 for Meteor */
#define METEOR_MTD_CFI_HIGH_START		0x00000001F0800000 
#define METEOR_MTD_CFI_HIGH_LEN			0x00800000
#define METEOR_MTD_CFI_HIGH_BUSWIDTH		(2)

/* CPLD information. These are register which control write access
   to the flash chips. */
#define METEOR_CPLD_BASE_ADDRESS		(long long) 0x00000001F8000000 
#define METEOR_CPLD_SIZE			0x20
#define METEOR_CPLD_CFI_WRITE_ENABLE_OFFSET	0x26
#define METEOR_CPLD_CFI_WRITE_ENABLE_BITS	0x00
#define METEOR_CPLD_JEDEC_WRITE_ENABLE_OFFSET	0x25
#define METEOR_CPLD_JEDEC_WRITE_ENABLE_BITS	0x01

/* -------------------------------------------------------------------------
 * Saturn
 * Flash information for meteor aka SWBD42 .
 * 4MB of boot flash of type JEDEC.
 * 64 MB of user flash of type CFI.
 *------------------------------------------------------------------------*/

/* JEDEC boot PROM */
# define SATURN_MTD_BOOT_CFI_START		0x00000001FFC00000 
# define SATURN_MTD_BOOT_CFI_LEN		0x00400000
# define SATURN_MTD_BOOT_CFI_BUSWIDTH		(2)

/* CFI bulk flash chip 1 for Saturn */
# define SATURN_MTD_CFI_LOW_START		0x00000001F0000000 
# define SATURN_MTD_CFI_LOW_LEN			0x00800000
# define SATURN_MTD_CFI_LOW_BUSWIDTH		(2)

/* CFI bulk flash chip 2 for Saturn */
#define SATURN_MTD_CFI_HIGH_START		0x00000001F0800000 
#define SATURN_MTD_CFI_HIGH_LEN			0x00800000
#define SATURN_MTD_CFI_HIGH_BUSWIDTH		(2)

/* CPLD information. These are register which control write access
   to the flash chips. */
#define SATURN_CPLD_BASE_ADDRESS		(long long) 0x00000001F8000000 
#define SATURN_CPLD_SIZE			0x20
#define SATURN_CPLD_CFI_WRITE_ENABLE_OFFSET	0x26
#define SATURN_CPLD_CFI_WRITE_ENABLE_BITS	0x03
#define SATURN_CPLD_JEDEC_WRITE_ENABLE_OFFSET	0x25
#define SATURN_CPLD_JEDEC_WRITE_ENABLE_BITS	0x01


/* -------------------------------------------------------------------------
 * pulsar and stealth and icecube
 * Flash information for pulsar aka silkworm ce-2.
 * 4  MB of boot flash of type CFI.
 * pulsar: 16 MB of user flash of type CFI, two seperate chips in 2 banks
 * of (1x16)
 * stealth: 16 MB of user flash of type CFI, one chips in 1 bank of (1x16)
 * icecube: no user flash of any type, boot flash only.
 *------------------------------------------------------------------------*/

/* CFI boot flash */
# define PULSAR_MTD_BOOT_CFI_START		0x00000001FFC00000 
# define PULSAR_MTD_BOOT_CFI_LEN		0x00400000
# define PULSAR_MTD_BOOT_CFI_BUSWIDTH		(2)


/* CFI bulk flash */
# define PULSAR_MTD_MAIN_LOW_CFI_START		0x00000001F0000000 
# define PULSAR_MTD_MAIN_LOW_CFI_LEN		0x00800000
# define PULSAR_MTD_MAIN_LOW_CFI_BUSWIDTH	(2)

/* CFI bulk flash */
# define PULSAR_MTD_MAIN_HIGH_CFI_START		0x00000001F0800000 
# define PULSAR_MTD_MAIN_HIGH_CFI_LEN		0x00800000
# define PULSAR_MTD_MAIN_HIGH_CFI_BUSWIDTH	(2)

/* CPLD information. These are register which control write access
   to the flash chips. */
#define PULSAR_CPLD_BASE_ADDRESS		0x00000001F8000000
#define PULSAR_CPLD_SIZE			0x20
#define PULSAR_CPLD_CFI_WRITE_ENABLE_OFFSET	0x94	/* main flash */
#define PULSAR_CPLD_CFI_WRITE_ENABLE_BITS	0x01
#define PULSAR_CPLD_BOOT_WRITE_ENABLE_OFFSET	0x90	/* boot cfi flash */
#define PULSAR_CPLD_BOOT_WRITE_ENABLE_BITS	0x01	

#define ELARA_CPLD_BOOT_WRITE_ENABLE_OFFSET 0x25    /* boot flash wp ctl */
#define ELARA_CPLD_BOOT_WRITE_ENABLE_BITS	0x01	

/* -------------------------------------------------------------------------
 * scimitar
 * Flash information for Freescale based scimitar
 * 4  MB of boot flash of type CFI. Rest is sam as Pulsar.
 *------------------------------------------------------------------------*/
# define SCIMITAR_MTD_BOOT_CFI_START		0xFFC00000 
# define SCIMITAR_MTD_BOOT_CFI_LEN		0x00400000
# define SCIMITAR_MTD_BOOT_CFI_BUSWIDTH		(2)

/* -------------------------------------------------------------------------
 * Lance 
 * Flash information for Freescale based Lance 
 * 16  MB of boot flash of type CFI. 
 *------------------------------------------------------------------------*/
# define LANCE_MTD_BOOT_CFI_START		0xFF800000 
# define LANCE_MTD_BOOT_CFI_LEN		0x0100000
# define LANCE_MTD_BOOT_CFI_BUSWIDTH		(2)

/* -------------------------------------------------------------------------
 * Marathon 
 * Flash information for Marathon
 * 4 MB of boot flash of type CFI.
 * pulsar: no user flash of any type, boot flash only. 
 *------------------------------------------------------------------------*/

/* CFI boot flash */
# define MARATHON_MTD_BOOT_CFI_START		0x00000001FFC00000ULL 
# define MARATHON_MTD_BOOT_CFI_LEN		0x00400000
# define MARATHON_MTD_BOOT_CFI_BUSWIDTH		(2)

/* CPLD information. 
 * These are register which control write access
 * to the flash chips. - Note the CPLD address is different for marathon
 */
#define MARATHON_CPLD_BASE_ADDRESS		0x00000001F8000000ULL
#define MARATHON_CPLD_SIZE			0x100
#define MARATHON_CPLD_BOOT_WRITE_ENABLE_OFFSET	0x00000040 /* boot cfi flash */
#define MARATHON_CPLD_BOOT_WRITE_ENABLE_BITS	0x00010000
#define MARATHON_CPLD_BOOT_WRITE_MAGIC		0xB1CA0000

/* -------------------------------------------------------------------------
 * athena
 * Flash information for athena aka silkworm ce-2.
 * 4  MB of boot flash of type CFI.
 * 16 MB of user flash of type CFI, two seperate chips in 2 banks of (1x16)
 *------------------------------------------------------------------------*/

/* CFI boot flash */
#define ATHENA_MTD_BOOT_CFI_START		0x00000001FFC00000 
#define ATHENA_MTD_BOOT_CFI_LEN			0x00400000
#define ATHENA_MTD_BOOT_CFI_BUSWIDTH		(2)

/* CFI bulk flash */
#define ATHENA_MTD_MAIN_LOW_CFI_START		0x00000001F0000000 
#define ATHENA_MTD_MAIN_LOW_CFI_LEN		0x00800000
#define ATHENA_MTD_MAIN_LOW_CFI_BUSWIDTH	(2)

/* CFI bulk flash */
//#define ATHENA_MTD_MAIN_HIGH_CFI_START	0x00000001F1000000 	//fixme when bootrom, fwdl, etc are updated to use MEM()0xf1000000
#define ATHENA_MTD_MAIN_HIGH_CFI_START		0x00000001F0800000 
#define ATHENA_MTD_MAIN_HIGH_CFI_LEN		0x00800000
#define ATHENA_MTD_MAIN_HIGH_CFI_BUSWIDTH	(2)

/* CPLD information. These are register which control write access
   to the flash chips. */
#define ATHENA_CPLD_BASE_ADDRESS		0x00000001F8000000
#define ATHENA_CPLD_SIZE			0x20
#define ATHENA_CPLD_CFI_WRITE_ENABLE_OFFSET	0x1C	/* main flash */
#define ATHENA_CPLD_CFI_WRITE_ENABLE_BITS	0x03
#define ATHENA_CPLD_BOOT_WRITE_ENABLE_OFFSET	0x1B	/* boot cfi flash */
#define ATHENA_CPLD_BOOT_WRITE_ENABLE_BITS	0x01	

/* -------------------------------------------------------------------------
 * Dazzler/Blazer and other low cost edition with similar maps
 * Flash information 
 * 4  MB of boot flash of type CFI.
 * 16 MB of user flash of type CFI, one chip in 1 bank (8MBx16)
 *------------------------------------------------------------------------*/

/* boot PROM */
/* 2MB Map - 512 KB on board -hopefully from start address */
# define LC_MTD_JEDEC_START	0xFFF80000
# define LC_MTD_JEDEC_LEN	0x00080000
# define LC_MTD_JEDEC_BUSWIDTH	(1)

/* CFI bulk flash */
# define LC_MTD_CFI_START	0xF0000000
# define LC_MTD_CFI_LEN		0x01000000
# define LC_MTD_CFI_BUSWIDTH	(2)

/* CPLD information. These are register which control write access
   to the flash chips. */
#define LC_CPLD_BASE_ADDRESS		    0xF8000000
#define LC_CPLD_SIZE			    0x100
#define LC_CPLD_JEDEC_WRITE_ENABLE_OFFSET   0x90
#define LC_CPLD_JEDEC_WRITE_ENABLE_BITS	    0x01
#define LC_CPLD_CFI_WRITE_ENABLE_OFFSET	    0x94
#define LC_CPLD_CFI_WRITE_ENABLE_BITS	    0x01

/* -------------------------------------------------------------------------
 * Forte Flash information:
 *  CS0: two Intel 28F128J3A in parallel, total 32 MB,
 *  CS1: two Intel 28F128J3A in parallel, total 32 MB.
 *------------------------------------------------------------------------*/

/* CFI flash CS0 */
# define FORTE_MTD_CFI_HIGH_START     0x00000001FE000000 
# define FORTE_MTD_CFI_HIGH_LEN       0x02000000
# define FORTE_MTD_CFI_HIGH_BUSWIDTH  (4)

/* CFI flash CS1 */
#define FORTE_MTD_CFI_LOW_START       0x00000001FC000000 
#define FORTE_MTD_CFI_LOW_LEN         0x02000000
#define FORTE_MTD_CFI_LOW_BUSWIDTH    (4)

/* CPLD information. These are register which control write access
   to the flash chips. */
#define FORTE_CPLD_BASE_ADDRESS			0x00000001F8000000
#define FORTE_CPLD_SIZE			0x100
#define FORTE_CPLD_CFI_WRITE_ENABLE_OFFSET	0x00000040
#define FORTE_CPLD_CFI_WRITE_ENABLE_BITS	0x00010000
#define FORTE_CPLD_BOOT_WRITE_ENABLE_OFFSET	FORTE_CPLD_CFI_WRITE_ENABLE_OFFSET
#define FORTE_CPLD_BOOT_WRITE_ENABLE_BITS	FORTE_CPLD_CFI_WRITE_ENABLE_BITS


map_word silkworm_read8(struct map_info *map, unsigned long ofs)
{
	map_word val;

	val.x[0] = __raw_readb((void __iomem *)(map->map_priv_1 + ofs));
	return val;
}

map_word silkworm_read16(struct map_info *map, unsigned long ofs)
{
	map_word val;

	val.x[0] =  __raw_readw((void __iomem *)(map->map_priv_1 + ofs));
	return val;
}

map_word silkworm_read32(struct map_info *map, unsigned long ofs)
{
	map_word val;

	val.x[0] =  __raw_readl((void __iomem *)(map->map_priv_1 + ofs));
	return val;
}

void silkworm_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, (void __iomem *)(map->map_priv_1 + from), len);
}

void silkworm_write8(struct map_info *map, const map_word d, unsigned long adr)
{
	__raw_writeb(d.x[0], (void __iomem *)(map->map_priv_1 + adr));
	mb();
}

void silkworm_write16(struct map_info *map, const map_word d, unsigned long adr)
{
	__raw_writew(d.x[0], (void __iomem *)(map->map_priv_1 + adr));
	mb();
}

void silkworm_write32(struct map_info *map, const map_word d, unsigned long adr)
{
	__raw_writel(d.x[0], (void __iomem *)(map->map_priv_1 + adr));
	mb();
}

void silkworm_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio((void __iomem *)(map->map_priv_1 + to), from, len);
}

void silkworm_set_cfi_vpp(struct map_info *map,int vpp)
{
	uint8_t *reg = cpld_addr + cfi_vpp_offset;
	if(vpp)
		*reg |= cfi_vpp_mask;
	else
		*reg &= ~cfi_vpp_mask;
	wmb();
}

void silkworm_set_jedec_vpp(struct map_info *map,int vpp)
{
	uint8_t *reg = cpld_addr + jedec_vpp_offset;
	if(vpp)
		*reg |= jedec_vpp_mask;
	else
		*reg &= ~jedec_vpp_mask;
	wmb();
}

void elara_set_jedec_vpp(struct map_info *map,int vpp)
{
	uint8_t *reg = cpld_addr + jedec_vpp_offset;
	if(vpp)
		*reg &= ~jedec_vpp_mask;
	else
		*reg |= jedec_vpp_mask;
	wmb();
}

/* 32 bit version of set VPP.  Note this is a negative assertion version. */
void silkworm_set_cfi_vpp32_n(struct map_info *map,int vpp)
{
	uint32_t *reg = (uint32_t *)(cpld_addr + cfi_vpp_offset32);
	if(vpp)
		*reg &= ~cfi_vpp_mask32;
	else
		*reg |= cfi_vpp_mask32;
	wmb();
}


/* 32 bit version of set VPP.  Note this is a negative assertion version. */
void silkworm_set_jedec_vpp32_n(struct map_info *map,int vpp)
{
	uint32_t *reg = (uint32_t *)(cpld_addr + jedec_vpp_offset32);
	if(vpp)
		*reg &= ~jedec_vpp_mask32 | jedec_vpp_magic32;
	else
		*reg |= jedec_vpp_mask32 | jedec_vpp_magic32;
	wmb();
}


/* -------------------------------------------------------------------------
 * First, the map and partitions for Ulysses.
 * Intel 28F128J3A in a 2 x 16 configuration 
 */
struct map_info ulysses_cfi_map = {
		name: 		"User flash",
		size: 		ULYSSES_MTD_CFI_LEN,
		bankwidth: 	ULYSSES_MTD_CFI_BUSWIDTH, /* 4 */
		read: 		silkworm_read32,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write32,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_cfi_vpp
};
struct mtd_partition ulysses_cfi_parts[] = {
	{
		name	: "Entire user flash",
		offset	: 0x00000000,
		size	: ULYSSES_MTD_CFI_LEN,
	},
	{
		name	: KERN0_STR "kernel and initrd (1)",
		offset	: 0x00000000,
		size	: 0x00400000,
	},
	{
		name	: KERN1_STR "kernel and initrd (2)",
		offset	: 0x00800000,
		size	: 0x00400000,
	},
	{
		name	: DUMP0_STR "log data (1)",
		offset	: 0x00400000,
		size	: 0x00400000,
	},
	{
		name	: DUMP1_STR "log data (2)",
		offset	: 0x00c00000,
		size	: 0x00400000,
	}
};
#define ULYSSES_CFI_NUM_PARTS	(elemof(ulysses_cfi_parts))
#define ULYSSES_CRASH_DUMP_PARTITION	3	/* index into ulysses_cfi_parts:  log data(1) */

/* AMD 29W040B in a 1 x 8 configuration */
struct map_info ulysses_jedec_map = {
		name: 		"Boot flash",
		size: 		ULYSSES_MTD_JEDEC_LEN,
		bankwidth: 	ULYSSES_MTD_JEDEC_BUSWIDTH, /* 1 */
		read: 		silkworm_read8,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write8,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_jedec_vpp
};
struct mtd_partition ulysses_jedec_parts[] = {
	{
		name	: BOOTENV_STR "boot environment",
		offset	: 0x00000000,
		size	: 0x00010000,
	},
	{
		name	: PROM_STR "boot prom",
		offset	: 0x00010000,
		size	: 0x00070000,
	}
};
#define ULYSSES_JEDEC_NUM_PARTS	(elemof(ulysses_jedec_parts))

struct mtd_partition pseudo_ulysses_jedec_parts[] = {
	{
		name	: PSEUDO_BOOTENV0_STR "boot environment",
		offset	: 0x00000000,
		size	: 0x00010000,
	},
	{
		name	: PSEUDO_BOOTENV1_STR "boot environment",
		offset	: 0x00010000,
		size	: 0x00010000,
	},
};

#define PSEUDO_ULYSSES_JEDEC_NUM_PARTS	(elemof(pseudo_ulysses_jedec_parts))

/* -------------------------------------------------------------------------
 * Next, the map and partitions for Warp.
 * Am29DL323D in 2x16 configuration 
 */
struct map_info warp_cfi_map = {
		name: 		"User flash",
		size: 		WARP_MTD_CFI_LEN,
		bankwidth: 	WARP_MTD_CFI_BUSWIDTH, /* 4 */
		read:	 	silkworm_read32,
		copy_from: 	silkworm_copy_from,
		write: 		silkworm_write32,
		copy_to: 	silkworm_copy_to,
		set_vpp:	NULL 	/* none for warp */
};
struct mtd_partition warp_cfi_parts[] = {
	{
		name:	"Entire user flash",
		offset:	0x00000000, // For diag purposes.
		size:	0x00400000, // 256 x (2 x 128k) byte sectors,
	},
	{
		name:	"big flash 128k sectors",
		offset:	0x00000000,
		size:	0x007E0000, // 63 x (2 x 64k) byte sectors
	},
	{
		name:	"big flash 16k sectors",
		offset:	0x007E0000,
		size:	0x00020000, // 8  x (2 x 8k)  byte sectors
	}
};
#define WARP_CFI_NUM_PARTS	(elemof(warp_cfi_parts))

/* AMD 29W040B in a 1 x 8 configuration */
struct map_info warp_jedec_map = {
		name: 		"Boot flash",
		size: 		WARP_MTD_JEDEC_LEN,
		bankwidth: 	WARP_MTD_JEDEC_BUSWIDTH, /* 1 */
		read: 		silkworm_read8,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write8,
		copy_to: 	silkworm_copy_to,
		set_vpp:	NULL	/* none for warp */
};
struct mtd_partition warp_jedec_parts[] = {
	{
		name	: BOOTENV_STR "boot environment",
		offset	: 0x00000000,
		size	: 0x00010000,
	},
	{
		name	: PROM_STR "boot prom",
		offset	: 0x00010000,
		size	: 0x00070000,
	}
};
#define WARP_JEDEC_NUM_PARTS	(elemof(warp_jedec_parts))


/* -------------------------------------------------------------------------
 * Last, the map and partitions for Terminator.
 * Intel 28F128J3A in a 2 x 16 configuration 
 */
struct map_info terminator_cfi_map = {
		name: 		"User flash",
		size: 		TERMINATOR_MTD_CFI_LEN,
		bankwidth: 	TERMINATOR_MTD_CFI_BUSWIDTH, /* 4 */
		read:	 	silkworm_read32,
		copy_from: 	silkworm_copy_from,
		write: 		silkworm_write32,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_cfi_vpp
};
struct mtd_partition terminator_cfi_parts[] = {
	{
		name	: "Entire user flash",
		offset	: 0x00000000,
		size	: TERMINATOR_MTD_CFI_LEN,
	},
	{
		name	: KERN0_STR "kernel and initrd (1)",
		offset	: 0x00000000,
		size	: 0x00400000,
	},
	{
		name	: KERN1_STR "kernel and initrd (2)",
		offset	: 0x00800000,
		size	: 0x00400000,
	},
	{
		name	: DUMP0_STR "log data (1)",
		offset	: 0x00400000,
		size	: 0x00400000,
	},
	{
		name	: DUMP1_STR "log data (2)",
		offset	: 0x00c00000,
		size	: 0x00400000,
	}
};
#define TERMINATOR_CFI_NUM_PARTS	(elemof(terminator_cfi_parts))
#define TERMINATOR_CRASH_DUMP_PARTITION	3	/* CFI partition log data(1) */

/* AMD 29W040B in a 1 x 8 configuration */
struct map_info terminator_jedec_map = {
		name: 		"Boot flash",
		size: 		TERMINATOR_MTD_JEDEC_LEN,
		bankwidth: 	TERMINATOR_MTD_JEDEC_BUSWIDTH, /* 1 */
		read: 		silkworm_read8,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write8,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_jedec_vpp
};
struct mtd_partition terminator_jedec_parts[] = {
	{
		name	: BOOTENV_STR "boot environment",
		offset	: 0x00000000,
		size	: 0x00010000,
	},
	{
		name	: PROM_STR "boot prom",
		offset	: 0x00010000,
		size	: 0x00070000,
	}
};
#define TERMINATOR_JEDEC_NUM_PARTS	(elemof(terminator_jedec_parts))

struct mtd_partition pseudo_terminator_jedec_parts[] = {
	{
		name	: PSEUDO_BOOTENV0_STR "boot environment",
		offset	: 0x00000000,
		size	: 0x00010000,
	},
	{
		name	: PSEUDO_BOOTENV1_STR "boot environment",
		offset	: 0x00010000,
		size	: 0x00010000,
	},
};
#define PSEUDO_TERMINATOR_JEDEC_NUM_PARTS	(elemof(pseudo_terminator_jedec_parts))

/* -------------------------------------------------------------------------
 * The map and partitions for Meteor.
 * There are two Intel 28F128J3A chips in a 1 x 16 configuration in Meteor.
 * These chips have separate chip selects. Thus, we need to have two map_info
 * sturctures, one for each chip. Similarly, we need have two mtd_partition
 * structures as well.
 */

struct map_info meteor_cfi_low_map = {
		name: 		"User flash",
		size: 		METEOR_MTD_CFI_LOW_LEN,
		bankwidth: 	METEOR_MTD_CFI_LOW_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_cfi_vpp
};

struct map_info meteor_cfi_high_map = {
		name: 		"User flash (second bank)",
		size: 		METEOR_MTD_CFI_HIGH_LEN,
		bankwidth: 	METEOR_MTD_CFI_HIGH_BUSWIDTH,	/* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write: 		silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_cfi_vpp
};

struct mtd_partition meteor_cfi_low_parts[] = {
	{
		name	: "User flash 1",
		offset	: 0x00000000,
		size	: METEOR_MTD_CFI_LOW_LEN,
	},
	{
		name	: KERN0_STR "kernel and initrd (1)",
		offset	: 0x00000000,
		size	: 0x00400000,
	},
	{
		name	: DUMP0_STR "log data (1)",
		offset	: 0x00400000,
		size	: 0x00400000,
	}
};

struct mtd_partition meteor_cfi_high_parts[] = {
	{
		name	: "User flash 2",
		offset	: 0x00000000,
		size	: METEOR_MTD_CFI_HIGH_LEN,
	},
	{
		name	: KERN1_STR "Kernel and initrd (2)",
		offset	: 0x00000000,
		size	: 0x00400000,
	},
	{
		name	: DUMP1_STR "log data (2)",
		offset	: 0x00400000,
		size	: 0x00400000,
	}
};

#define METEOR_CFI_LOW_NUM_PARTS	(elemof(meteor_cfi_low_parts))
#define METEOR_CFI_HIGH_NUM_PARTS   	(elemof(meteor_cfi_high_parts))

#define METEOR_CRASH_DUMP_PARTITION	2	/* CFI partition log data(1) */

/* AMD 29W040B in a 1 x 8 configuration */
struct map_info meteor_jedec_map = {
		name: 		"Boot flash",
		size: 		METEOR_MTD_JEDEC_LEN,
		bankwidth: 	METEOR_MTD_JEDEC_BUSWIDTH, /* 1 */
		read: 		silkworm_read8,
		copy_from: 	silkworm_copy_from,
		write: 		silkworm_write8,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_jedec_vpp
};
struct mtd_partition meteor_jedec_parts[] = {
	{
		name	: BOOTENV_STR "boot environment",
		offset	: 0x00000000,
		size	: 0x00010000,
	},
	{
		name	: PROM_STR "boot prom",
		offset	: 0x00020000,
		size	: 0x00060000,
	}
};

#define METEOR_JEDEC_NUM_PARTS (elemof(meteor_jedec_parts))

struct mtd_partition pseudo_meteor_jedec_parts[] = {
	{
		name	: PSEUDO_BOOTENV0_STR "boot environment",
		offset	: 0x00000000,
		size	: 0x00010000,
	},
	{
		name	: PSEUDO_BOOTENV1_STR "boot environment",
		offset	: 0x00010000,
		size	: 0x00010000,
	},
};
#define PSEUDO_METEOR_JEDEC_NUM_PARTS	(elemof(pseudo_meteor_jedec_parts))

/* -------------------------------------------------------------------------
 * The map and partitions for Saturn.
 * User Flash - CFI 
 * 	Total memory 16MB
 * Boot Flash - CFI
 *	Total 4MB	
 */

struct map_info saturn_cfi_low_map = {
		name: 		"User flash",
		size: 		SATURN_MTD_CFI_LOW_LEN,
		bankwidth: 	SATURN_MTD_CFI_LOW_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_cfi_vpp
};

struct map_info saturn_cfi_high_map = {
		name: 		"User flash (second bank)",
		size: 		SATURN_MTD_CFI_HIGH_LEN,
		bankwidth: 	SATURN_MTD_CFI_HIGH_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write: 		silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_cfi_vpp
};

struct mtd_partition saturn_cfi_low_parts[] = {
	{
		name	: "User flash 1",
		offset	: 0x00000000,
		size	: SATURN_MTD_CFI_LOW_LEN,
	},
	{
		name	: KERN0_STR "kernel and initrd (1)",
		offset	: 0x00000000,
		size	: 0x00400000,
	},
	{
		name	: DUMP0_STR "log data (1)",
		offset	: 0x00400000,
		size	: 0x00400000,
	}
};

struct mtd_partition saturn_cfi_high_parts[] = {
	{
		name	: "User flash 2",
		offset	: 0x00000000,
		size	: SATURN_MTD_CFI_HIGH_LEN,
	},
	{
		name	: KERN1_STR "Kernel and initrd (2)",
		offset	: 0x00000000,
		size	: 0x00400000,
	},
	{
		name	: DUMP1_STR "log data (2)",
		offset	: 0x00400000,
		size	: 0x00400000,
	}
};

#define SATURN_CFI_LOW_NUM_PARTS	(elemof(meteor_cfi_low_parts))
#define SATURN_CFI_HIGH_NUM_PARTS   	(elemof(meteor_cfi_high_parts))

#define SATURN_CRASH_DUMP_PARTITION	2	/* CFI partition log data(1) */

struct map_info saturn_cfi_boot_map = {
		name: 		"Boot flash",
		size: 		SATURN_MTD_BOOT_CFI_LEN,
		bankwidth: 	SATURN_MTD_BOOT_CFI_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_jedec_vpp
};
struct mtd_partition saturn_cfi_boot_parts[] = {
	{
		name	: BOOTENV0_STR "boot environment (1)",
		offset	: 0x00000000,
		size	: 0x00020000,
	},
	{
		name	: BOOTENV1_STR "boot environment (2)",
		offset	: 0x00020000,
		size	: 0x00020000,
	},
	{
		name	: PROM0_STR "boot prom (1)",
		offset	: 0x00040000,
		size	: 0x001c0000,
	},
	{
		name	: PROM1_STR "boot prom (2)",
		offset	: 0x00200000,
		size	: 0x001c0000,
	},
	{
		name	: "unused",
		offset	: 0x003c0000,
		size	: 0x00020000,
	},
	{
		name	: BOOTSEL_STR "boot prom selector",
		offset	: 0x003e0000,
		size	: 0x00020000,
	},
};
#define SATURN_JEDEC_NUM_PARTS	(elemof(saturn_cfi_boot_parts))

/* -------------------------------------------------------------------------
 * The map and partitions for Pulsar and stealth.
 * Pulsar: Main flash: 2 banks Intel ?? in 2 bank x (1x16) configuration.
 * Stealth: Main flash: 1 banks Intel ?? in 1 bank x (1x16) configuration.
 * Boot flash: 1 bank Intel ?? in a (1x16) configuration.
 */
struct map_info pulsar_cfi_main_low_map= {
		name: 		"User flash low",
		size: 		PULSAR_MTD_MAIN_LOW_CFI_LEN,
		bankwidth: 	PULSAR_MTD_MAIN_LOW_CFI_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_cfi_vpp
};

struct map_info pulsar_cfi_main_high_map = {
		name: 		"User flash high",
		size: 		PULSAR_MTD_MAIN_HIGH_CFI_LEN,
		bankwidth: 	PULSAR_MTD_MAIN_HIGH_CFI_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write: 		silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_cfi_vpp
};
struct mtd_partition pulsar_cfi_main_low_parts[] = {
	{
		name	: "Entire user flash (first bank)",
		offset	: 0x00000000,
		size	: PULSAR_MTD_MAIN_LOW_CFI_LEN,
	},
	{
		name	: KERN0_STR "kernel and initrd (1)",
		offset	: 0x00000000,
		size	: 0x00400000,
	},
	{
		name	: DUMP0_STR "log data (1)",
		offset	: 0x00400000,
		size	: 0x00400000,
	}
};

struct mtd_partition pulsar_cfi_main_high_parts[] = {
	{
		name	: "Entire user flash (second bank)",
		offset	: 0x00000000,
		size	: PULSAR_MTD_MAIN_HIGH_CFI_LEN,
	},
	{
		name	: KERN1_STR "kernel and initrd (2)",
		offset	: 0x00000000,
		size	: 0x00400000,
	},
	{
		name	: DUMP1_STR "log data (2)",
		offset	: 0x00400000,
		size	: 0x00400000,
	}
};
#define PULSAR_CFI_MAIN_LOW_NUM_PARTS	(elemof(pulsar_cfi_main_low_parts))
#define PULSAR_CFI_MAIN_HIGH_NUM_PARTS	(elemof(pulsar_cfi_main_high_parts))
#define PULSAR_CRASH_DUMP_PARTITION	2	/* CFI partition log data(1) */

struct map_info pulsar_cfi_boot_map = {
		name: 		"Boot flash",
		size: 		PULSAR_MTD_BOOT_CFI_LEN,
		bankwidth: 	PULSAR_MTD_BOOT_CFI_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write: 		silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_jedec_vpp	/* not really jedec, just using this fn */
};
struct mtd_partition pulsar_cfi_boot_parts[] = {
	{
		name	: BOOTENV0_STR "boot environment (1)",
		offset	: 0x00000000,
		size	: 0x00020000,
	},
	{
		name	: BOOTENV1_STR "boot environment (2)",
		offset	: 0x00020000,
		size	: 0x00020000,
	},
	{
		name	: PROM0_STR "boot prom (1)",
		offset	: 0x00040000,
		size	: 0x001c0000,
	},
	{
		name	: PROM1_STR "boot prom (2)",
		offset	: 0x00200000,
		size	: 0x001c0000,
	},
	{
		name	: "unused",
		offset	: 0x003c0000,
		size	: 0x00020000,
	},
	{
		name	: BOOTSEL_STR "boot prom selector",
		offset	: 0x003e0000,
		size	: 0x00020000,
	},
};

#define PULSAR_CFI_BOOT_NUM_PARTS	(elemof(pulsar_cfi_boot_parts))

struct map_info scimitar_cfi_boot_map = {
		name: 		"Boot flash",
		size: 		SCIMITAR_MTD_BOOT_CFI_LEN,
		bankwidth: 	SCIMITAR_MTD_BOOT_CFI_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write: 		silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_jedec_vpp	/* not really jedec, just using this fn */
};
struct mtd_partition scimitar_cfi_boot_parts[] = {
    {
        name    : BOOTENV1_STR "boot environment (1)",
        offset  : 0x00000000,
        size    : 0x00020000,
    },
    {
        name    : BOOTENV0_STR "boot environment (2)",
        offset  : 0x00020000,
        size    : 0x00020000,
    },
    {
        name    : PROM0_STR "boot prom (1)",
        offset  : 0x00040000,
        size    : 0x001c0000,
    },
    {
        name    : PROM1_STR "boot prom (2)",
        offset  : 0x00200000,
        size    : 0x001c0000,
    },
    {
        name    : "unused",
        offset  : 0x003c0000,
        size    : 0x00020000,
    },
    {
        name    : BOOTSEL_STR "boot prom selector",
        offset  : 0x003e0000,
        size    : 0x00020000,
    },
};
#define SCIMITAR_CFI_BOOT_NUM_PARTS	(elemof(scimitar_cfi_boot_parts))
struct mtd_partition zentron_cfi_boot_parts[] = {
	{
		name	: BOOTENV1_STR "boot environment (1)",
		offset	: 0x00000000,
		size	: 0x00020000,
	},
	{
		name	: BOOTENV0_STR "boot environment (2)",
		offset	: 0x00020000,
		size	: 0x00020000,
	},
	{
		name	: PROM0_STR "boot prom (1)",
		offset	: 0x00040000,
		size	: 0x001c0000,
	},
	{
		name	: PROM1_STR "boot prom (2)",
		offset	: 0x00200000,
		size	: 0x001c0000,
	},
	{
		name	: "unused",
		offset	: 0x003c0000,
		size	: 0x00020000,
	},
	{
		name	: BOOTSEL_STR "boot prom selector",
		offset	: 0x003e0000,
		size	: 0x00020000,
	},
};
#define ZENTRON_CFI_BOOT_NUM_PARTS	(elemof(zentron_cfi_boot_parts))

struct map_info elara_cfi_boot_map = {
		name: 		"Boot flash",
		size: 		SCIMITAR_MTD_BOOT_CFI_LEN,
		bankwidth: 	SCIMITAR_MTD_BOOT_CFI_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write: 		silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	elara_set_jedec_vpp	/* not really jedec, just using this fn */
};

/* -------------------------------------------------------------------------
 * The map and partitions for Blazer/Dazzler - other LC (Low Cost) platforms.
 */
struct map_info lc_cfi_map = {
		name: 		"User flash",
		size: 		LC_MTD_CFI_LEN,
		bankwidth: 	LC_MTD_CFI_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_cfi_vpp
};
struct mtd_partition lc_cfi_parts[] = {
	{
		name	: "Entire user flash",
		offset	: 0x00000000,
		size	: LC_MTD_CFI_LEN,
	},
	{
		name	: KERN0_STR "kernel and initrd (1)",
		offset	: 0x00000000,
		size	: 0x00400000,
	},
	{
		name	: KERN1_STR "kernel and initrd (2)",
		offset	: 0x00800000,
		size	: 0x00400000,
	},
	{
		name	: DUMP0_STR "log data (1)",
		offset	: 0x00400000,
		size	: 0x00400000,
	},
	{
		name	: DUMP1_STR "log data (2)",
		offset	: 0x00c00000,
		size	: 0x00400000,
	}
};

#define LC_CFI_NUM_PARTS	(elemof(lc_cfi_parts))
#define LC_CRASH_DUMP_PARTITION	3	/* CFI partition log data(1) */

/* AMD 29W040B in a 1 x 8 configuration */
struct map_info lc_jedec_map = {
		name: 		"Boot flash",
		size: 		LC_MTD_JEDEC_LEN,
		bankwidth: 	LC_MTD_JEDEC_BUSWIDTH, /* 1 */
		read: 		silkworm_read8,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write8,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_jedec_vpp
};
struct mtd_partition lc_jedec_parts[] = {
	{
		name	: BOOTENV_STR "boot environment",
		offset	: 0x00000000,
		size	: 0x00010000,
	},
	{
		name	: PROM_STR "boot prom",
		offset	: 0x00010000,
		size	: 0x00070000,
	}
};
#define LC_JEDEC_NUM_PARTS	(elemof(lc_jedec_parts))

struct mtd_partition pseudo_lc_jedec_parts[] = {
	{
		name	: PSEUDO_BOOTENV0_STR "boot environment",
		offset	: 0x00000000,
		size	: 0x00010000,
	},
	{
		name	: PSEUDO_BOOTENV1_STR "boot environment",
		offset	: 0x00010000,
		size	: 0x00010000,
	},
};

#define PSEUDO_LC_JEDEC_NUM_PARTS	(elemof(pseudo_lc_jedec_parts))

/* -------------------------------------------------------------------------
 * The map and partitions for Marathon.
 * Boot flash: 1 bank Intel ?? in a (1x16) configuration.
 */
struct map_info marathon_cfi_boot_map = {
		name: 		"Boot flash",
		size: 		MARATHON_MTD_BOOT_CFI_LEN,
		bankwidth: 	MARATHON_MTD_BOOT_CFI_BUSWIDTH,
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_jedec_vpp32_n	
		/* not really jedec, just using this fn */
};

struct mtd_partition marathon_cfi_boot_parts[] = {
	{
		name	: BOOTENV0_STR "boot environment (1)",
		offset	: 0x00000000,
		size	: 0x00020000,
	},
	{
		name	: BOOTENV1_STR "boot environment (2)",
		offset	: 0x00020000,
		size	: 0x00020000,
	},
	{
		name	: PROM0_STR "boot prom (1)",
		offset	: 0x00040000,
		size	: 0x001c0000,
	},
	{
		name	: PROM1_STR "boot prom (2)",
		offset	: 0x00200000,
		size	: 0x001c0000,
	},
	{
		name	: "unused",
		offset	: 0x003c0000,
		size	: 0x00020000,
	},
	{
		name	: BOOTSEL_STR "boot prom selector",
		offset	: 0x003e0000,
		size	: 0x00020000,
	},
};
#define MARATHON_CFI_BOOT_NUM_PARTS	(elemof(marathon_cfi_boot_parts))

/* -------------------------------------------------------------------------
 * The map and partitions for Athena.
 * Main flash: 2 banks Intel ?? in 2 bank x (1x16) configuration.
 * Boot flash: 1 bank Intel ?? in a (1x16) configuration.
 */
struct map_info athena_cfi_main_low_map= {
		name: 		"User flash low",
		size: 		ATHENA_MTD_MAIN_LOW_CFI_LEN,
		bankwidth: 	ATHENA_MTD_MAIN_LOW_CFI_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_cfi_vpp
};
struct map_info athena_cfi_main_high_map = {
		name: 		"User flash high",
		size: 		ATHENA_MTD_MAIN_HIGH_CFI_LEN,
		bankwidth: 	ATHENA_MTD_MAIN_HIGH_CFI_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_cfi_vpp
};
struct mtd_partition athena_cfi_main_low_parts[] = {
	{
		name	: "Entire user flash (first bank)",
		offset	: 0x00000000,
		size	: ATHENA_MTD_MAIN_LOW_CFI_LEN,
	},
	{
		name	: KERN0_STR "kernel and initrd (1)",
		offset	: 0x00000000,
		size	: 0x00400000,
	},
	{
		name	: DUMP0_STR "log data (1)",
		offset	: 0x00400000,
		size	: 0x00400000,
	}
};
struct mtd_partition athena_cfi_main_high_parts[] = {
	{
		name	: "Entire user flash (second bank)",
		offset	: 0x00000000,
		size	: ATHENA_MTD_MAIN_HIGH_CFI_LEN,
	},
	{
		name	: KERN1_STR "kernel and initrd (2)",
		offset	: 0x00000000,
		size	: 0x00400000,
	},
	{
		name	: DUMP1_STR "log data (2)",
		offset	: 0x00400000,
		size	: 0x00400000,
	}
};
#define ATHENA_CFI_MAIN_LOW_NUM_PARTS	(elemof(athena_cfi_main_low_parts))
#define ATHENA_CFI_MAIN_HIGH_NUM_PARTS	(elemof(athena_cfi_main_high_parts))
#define ATHENA_CRASH_DUMP_PARTITION	2	/* CFI partition log data(1) */

struct map_info athena_cfi_boot_map = {
		name: 		"Boot flash",
		size: 		ATHENA_MTD_BOOT_CFI_LEN,
		bankwidth: 	ATHENA_MTD_BOOT_CFI_BUSWIDTH, /* 2 */
		read:	 	silkworm_read16,
		copy_from: 	silkworm_copy_from,
		write:	 	silkworm_write16,
		copy_to: 	silkworm_copy_to,
		set_vpp:	silkworm_set_jedec_vpp	/* not really jedec, just using this fn */
};
struct mtd_partition athena_cfi_boot_parts[] = {
	{
		name	: BOOTENV0_STR "boot environment (1)",
		offset	: 0x00000000,
		size	: 0x00020000,
	},
	{
		name	: BOOTENV1_STR "boot environment (2)",
		offset	: 0x00020000,
		size	: 0x00020000,
	},
	{
		name	: PROM0_STR "boot prom (1)",
		offset	: 0x00040000,
		size	: 0x001c0000,
	},
	{
		name	: PROM1_STR "boot prom (2)",
		offset	: 0x00200000,
		size	: 0x001c0000,
	},
	{
		name	: "unused",
		offset	: 0x003c0000,
		size	: 0x00020000,
	},
	{
		name	: BOOTSEL_STR "boot prom selector",
		offset	: 0x003e0000,
		size	: 0x00020000,
	},
};
#define ATHENA_CFI_BOOT_NUM_PARTS	(elemof(athena_cfi_boot_parts))


/* -------------------------------------------------------------------------
 * Maps and partitions for Forte.
 */
struct map_info forte_cfi_low_map = {
    name:       "User flash 1 (low)",
    size:       FORTE_MTD_CFI_LOW_LEN,
    bankwidth:	FORTE_MTD_CFI_LOW_BUSWIDTH, /* 4 */
    read:       silkworm_read32,
    copy_from:  silkworm_copy_from,
    write:      silkworm_write32,
    copy_to:    silkworm_copy_to,
    set_vpp:    silkworm_set_cfi_vpp32_n
};

struct map_info forte_cfi_high_map = {
    name:       "User flash 2 (high)",
    size:       FORTE_MTD_CFI_HIGH_LEN,
    bankwidth:  FORTE_MTD_CFI_HIGH_BUSWIDTH, /* 4 */
    read:       silkworm_read32,
    copy_from:  silkworm_copy_from,
    write:      silkworm_write32,
    copy_to:    silkworm_copy_to,
    set_vpp:    silkworm_set_cfi_vpp32_n
};

struct mtd_partition forte_cfi_low_parts[] = {
  {
    name    : "user flash 1 (low)",
    offset  : 0x00000000,
    size    : FORTE_MTD_CFI_LOW_LEN,
  },
  {
    name    : KERN0_STR "Kernel and initrd (0)",
    offset  : 0x00000000,
    size    : 0x01C00000,
  },
  {
    name    : DUMP0_STR "Ras log (0)",
    offset  : 0x01C00000,
    size    : 0x00400000,
  }
};

struct mtd_partition forte_cfi_high_parts[] = {
  {
    name    : "user flash 2 (high)",
    offset  : 0x00000000,
    size    : FORTE_MTD_CFI_HIGH_LEN,
  },
  {
    name    : KERN1_STR "Kernel and initrd (1)",
    offset  : 0x00000000,
    size    : 0x01C00000,
  },
  {
    name    : DUMP1_STR "Ras log (1)",
    offset  : 0x01C00000,
    size    : 0x00100000,
  },
  {
    name    : "unused",
    offset  : 0x01D00000,
    size    : 0x00080000,
  },
  {
    name    : BOOTENV0_STR "Boot Environment (0)",
    offset  : 0x01D80000,
    size    : 0x00040000,
  },
  {
    name    : BOOTENV1_STR "Boot Environment (1)",
    offset  : 0x01DC0000,
    size    : 0x00040000,
  },
  {
    name    : PROM_STR "Boot Rom",
    offset  : 0x01E00000,
    size    : 0x00200000,
  },
};

#define FORTE_CFI_LOW_NUM_PARTS     (elemof(forte_cfi_low_parts))
#define FORTE_CFI_HIGH_NUM_PARTS    (elemof(forte_cfi_high_parts))
#define FORTE_CRASH_DUMP_PARTITION  2


/* -------------------------------------------------------------------------
 */

int __init init_silkworm(void)
{	
	int idx, ret = 0;
	int crashdump_part = 0;
	struct my_mtd_mapping_t *crashdump_map = NULL;
	int pulsar32 = 1;

	silkworm_maps = NULL;

	/* Determine our board type at runtime */
	switch(silkworm.board)
	{
		default:
			printk(KERN_NOTICE "silkworm: "
				"Unknown board type "
				"Using SWBD10 flash configuration\n");
			/* Fall through */
		case SWBD10:	
			printk(KERN_NOTICE "silkworm: "
				"Using SWBD10 flash configuration\n");
			/* Two separate mappings, one NULL terminator */
			if(!(silkworm_maps = kmalloc(sizeof(struct my_mtd_mapping_t) * 4, GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, sizeof(struct my_mtd_mapping_t) * 4);
			silkworm_maps[0].map	= &ulysses_jedec_map;
			silkworm_maps[0].parts	= &ulysses_jedec_parts[0];
			silkworm_maps[0].nparts	= ULYSSES_JEDEC_NUM_PARTS;
			silkworm_maps[0].probe	= "jedec";
			silkworm_maps[1].map	= &ulysses_cfi_map;
			silkworm_maps[1].parts	= &ulysses_cfi_parts[0];
			silkworm_maps[1].nparts	= ULYSSES_CFI_NUM_PARTS;
			silkworm_maps[1].probe	= "cfi_probe";
			silkworm_maps[2].map	= &ulysses_jedec_map;
			silkworm_maps[2].parts	= &pseudo_ulysses_jedec_parts[0];
			silkworm_maps[2].nparts	= PSEUDO_ULYSSES_JEDEC_NUM_PARTS;
			silkworm_maps[2].probe	= "jedec";
			silkworm_maps[0].base = (void *)ioremap_nocache(
					ULYSSES_MTD_JEDEC_START,
					ULYSSES_MTD_JEDEC_LEN);
			silkworm_maps[1].base = (void *)ioremap_nocache(
					ULYSSES_MTD_CFI_START,
					ULYSSES_MTD_CFI_LEN);
			silkworm_maps[2].base = silkworm_maps[0].base;
			cpld_addr = ioremap_nocache(
					ULYSSES_CPLD_BASE_ADDRESS,
					ULYSSES_CPLD_SIZE);
			cfi_vpp_offset = ULYSSES_CPLD_CFI_WRITE_ENABLE_OFFSET;
			cfi_vpp_mask = ULYSSES_CPLD_CFI_WRITE_ENABLE_BITS;
			jedec_vpp_offset = 
				ULYSSES_CPLD_JEDEC_WRITE_ENABLE_OFFSET;
			jedec_vpp_mask = 
				ULYSSES_CPLD_JEDEC_WRITE_ENABLE_BITS;
			crashdump_map = &silkworm_maps[1];
			crashdump_part = ULYSSES_CRASH_DUMP_PARTITION;
			break;
		case SWBD11:	
			printk(KERN_NOTICE "silkworm: "
				"SWBD11 flash configuration is NOT currently supported\n");
			break;
		case SWBD12:
		case SWBD31:
			printk(KERN_NOTICE "silkworm: "
				"Using SWBD12 flash configuration\n");

			/* Two separate mappings, one NULL terminator */
			if(!(silkworm_maps = kmalloc(sizeof(struct my_mtd_mapping_t) * 4, GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, sizeof(struct my_mtd_mapping_t) * 4);
			silkworm_maps[0].map	= &terminator_jedec_map;
			silkworm_maps[0].parts	= &terminator_jedec_parts[0];
			silkworm_maps[0].nparts	= TERMINATOR_JEDEC_NUM_PARTS;
			silkworm_maps[0].probe	= "jedec";
			silkworm_maps[1].map	= &terminator_cfi_map;
			silkworm_maps[1].parts	= &terminator_cfi_parts[0];
			silkworm_maps[1].nparts	= TERMINATOR_CFI_NUM_PARTS;
			silkworm_maps[1].probe	= "cfi_probe";
			silkworm_maps[2].map	= &terminator_jedec_map;
			silkworm_maps[2].parts	= &pseudo_terminator_jedec_parts[0];
			silkworm_maps[2].nparts	= PSEUDO_TERMINATOR_JEDEC_NUM_PARTS;
			silkworm_maps[2].probe	= "jedec";
			silkworm_maps[0].base = (void *)ioremap_nocache(
					TERMINATOR_MTD_JEDEC_START,
					TERMINATOR_MTD_JEDEC_LEN);
			silkworm_maps[1].base = (void *)ioremap_nocache(
					TERMINATOR_MTD_CFI_START,
					TERMINATOR_MTD_CFI_LEN);
			silkworm_maps[2].base = silkworm_maps[0].base;
			cpld_addr = ioremap_nocache(
					TERMINATOR_CPLD_BASE_ADDRESS,
					TERMINATOR_CPLD_SIZE);
			cfi_vpp_offset = 
				TERMINATOR_CPLD_CFI_WRITE_ENABLE_OFFSET;
			cfi_vpp_mask = 
				TERMINATOR_CPLD_CFI_WRITE_ENABLE_BITS;
			jedec_vpp_offset = 
				TERMINATOR_CPLD_JEDEC_WRITE_ENABLE_OFFSET;
			jedec_vpp_mask = 
				TERMINATOR_CPLD_JEDEC_WRITE_ENABLE_BITS;
			crashdump_map = &silkworm_maps[1];
			crashdump_part = TERMINATOR_CRASH_DUMP_PARTITION;
			break;
		/* Blazer/Dazzler and other similar lowcost families */
		case SWBD22:	
		case SWBD26:	
		case SWBD27:	
		case SWBD33:	
			printk(KERN_NOTICE "silkworm: "
				"Using SWBD%d flash configuration\n",silkworm.board);

			/* Two separate mappings, one NULL terminator */
			if(!(silkworm_maps = kmalloc(sizeof(struct my_mtd_mapping_t) * 4, GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, sizeof(struct my_mtd_mapping_t) * 4);
			silkworm_maps[0].map	= &lc_jedec_map;
			silkworm_maps[0].parts	= &lc_jedec_parts[0];
			silkworm_maps[0].nparts	= LC_JEDEC_NUM_PARTS;
			silkworm_maps[0].probe	= "jedec";
			silkworm_maps[1].map	= &lc_cfi_map;
			silkworm_maps[1].parts	= &lc_cfi_parts[0];
			silkworm_maps[1].nparts	= LC_CFI_NUM_PARTS;
			silkworm_maps[1].probe	= "cfi_probe";
			silkworm_maps[2].map	= &lc_jedec_map;
			silkworm_maps[2].parts	= &pseudo_lc_jedec_parts[0];
			silkworm_maps[2].nparts	= PSEUDO_LC_JEDEC_NUM_PARTS;
			silkworm_maps[2].probe	= "jedec";
			silkworm_maps[0].base = (void *)ioremap_nocache(
					LC_MTD_JEDEC_START,
					LC_MTD_JEDEC_LEN);
			silkworm_maps[1].base = (void *)ioremap_nocache(
					LC_MTD_CFI_START,
					LC_MTD_CFI_LEN);
			silkworm_maps[2].base = silkworm_maps[0].base;
			cpld_addr = ioremap_nocache(
					LC_CPLD_BASE_ADDRESS,
					LC_CPLD_SIZE);
			cfi_vpp_offset = 
				LC_CPLD_CFI_WRITE_ENABLE_OFFSET;
			cfi_vpp_mask = 
				LC_CPLD_CFI_WRITE_ENABLE_BITS;
			jedec_vpp_offset = 
				LC_CPLD_JEDEC_WRITE_ENABLE_OFFSET;
			jedec_vpp_mask = 
				LC_CPLD_JEDEC_WRITE_ENABLE_BITS;
			crashdump_map = &silkworm_maps[1];
			crashdump_part = LC_CRASH_DUMP_PARTITION;
			break;
		case SWBD21:	
			printk(KERN_NOTICE "silkworm: "
				"Using SWBD21 flash configuration\n");
			/* Three separate mappings, one NULL terminator */
			if(!(silkworm_maps = kmalloc(sizeof(struct my_mtd_mapping_t) * 5, GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, sizeof(struct my_mtd_mapping_t) * 5);
			silkworm_maps[0].map	= &meteor_jedec_map;
			silkworm_maps[0].parts	= &meteor_jedec_parts[0];
			silkworm_maps[0].nparts	= METEOR_JEDEC_NUM_PARTS;
			silkworm_maps[0].probe	= "jedec";
			silkworm_maps[1].map	= &meteor_cfi_low_map;
			silkworm_maps[1].parts	= &meteor_cfi_low_parts[0];
			silkworm_maps[1].nparts	= METEOR_CFI_LOW_NUM_PARTS;
			silkworm_maps[1].probe	= "cfi_probe";
			silkworm_maps[2].map	= &meteor_cfi_high_map;
			silkworm_maps[2].parts	= &meteor_cfi_high_parts[0];
			silkworm_maps[2].nparts	= METEOR_CFI_HIGH_NUM_PARTS;
			silkworm_maps[2].probe	= "cfi_probe";
			silkworm_maps[3].map	= &meteor_jedec_map;
			silkworm_maps[3].parts	= &pseudo_meteor_jedec_parts[0];
			silkworm_maps[3].nparts	= PSEUDO_METEOR_JEDEC_NUM_PARTS;
			silkworm_maps[3].probe	= "jedec";
			silkworm_maps[0].base = (void *)ioremap_nocache(
					METEOR_MTD_JEDEC_START,
					METEOR_MTD_JEDEC_LEN);
			silkworm_maps[1].base = (void *)ioremap_nocache(
					METEOR_MTD_CFI_LOW_START,
					METEOR_MTD_CFI_LOW_LEN);
			silkworm_maps[2].base = (void *)ioremap_nocache(
					METEOR_MTD_CFI_HIGH_START,
					METEOR_MTD_CFI_HIGH_LEN);
			silkworm_maps[3].base = silkworm_maps[0].base;
			cpld_addr = ioremap_nocache(
					METEOR_CPLD_BASE_ADDRESS,
					METEOR_CPLD_SIZE);
			cfi_vpp_offset = METEOR_CPLD_CFI_WRITE_ENABLE_OFFSET;
			cfi_vpp_mask = METEOR_CPLD_CFI_WRITE_ENABLE_BITS;
			jedec_vpp_offset = 
				METEOR_CPLD_JEDEC_WRITE_ENABLE_OFFSET;
			jedec_vpp_mask = 
				METEOR_CPLD_JEDEC_WRITE_ENABLE_BITS;
			crashdump_map = &silkworm_maps[1];	/* use meteor_cfi_low_map */
			crashdump_part = METEOR_CRASH_DUMP_PARTITION;
			break;
		case SWBD42:	
			printk(KERN_NOTICE "silkworm: "
				"Using SWBD42 flash configuration\n");
			/* Three separate mappings, one NULL terminator */
			if(!(silkworm_maps = kmalloc(
				sizeof(struct my_mtd_mapping_t) * 4, 
					GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, 
				sizeof(struct my_mtd_mapping_t) * 4);
			silkworm_maps[0].map	= &saturn_cfi_boot_map;
			silkworm_maps[0].parts	= &saturn_cfi_boot_parts[0];
			silkworm_maps[0].nparts	= SATURN_JEDEC_NUM_PARTS;
			silkworm_maps[0].probe	= "cfi_probe";
			silkworm_maps[1].map	= &saturn_cfi_low_map;
			silkworm_maps[1].parts	= &saturn_cfi_low_parts[0];
			silkworm_maps[1].nparts	= SATURN_CFI_LOW_NUM_PARTS;
			silkworm_maps[1].probe	= "cfi_probe";
			silkworm_maps[2].map	= &saturn_cfi_high_map;
			silkworm_maps[2].parts	= &saturn_cfi_high_parts[0];
			silkworm_maps[2].nparts	= SATURN_CFI_HIGH_NUM_PARTS;
			silkworm_maps[2].probe	= "cfi_probe";
			silkworm_maps[0].base = (void *)ioremap_nocache(
					SATURN_MTD_BOOT_CFI_START,
					SATURN_MTD_BOOT_CFI_LEN);
			silkworm_maps[1].base = (void *)ioremap_nocache(
					SATURN_MTD_CFI_LOW_START,
					SATURN_MTD_CFI_LOW_LEN);
			silkworm_maps[2].base = (void *)ioremap_nocache(
					SATURN_MTD_CFI_HIGH_START,
					SATURN_MTD_CFI_HIGH_LEN);
			cpld_addr = ioremap_nocache(
					SATURN_CPLD_BASE_ADDRESS,
					SATURN_CPLD_SIZE);
			cfi_vpp_offset = SATURN_CPLD_CFI_WRITE_ENABLE_OFFSET;
			cfi_vpp_mask = SATURN_CPLD_CFI_WRITE_ENABLE_BITS;
			jedec_vpp_offset = 
				SATURN_CPLD_JEDEC_WRITE_ENABLE_OFFSET;
			jedec_vpp_mask = 
				SATURN_CPLD_JEDEC_WRITE_ENABLE_BITS;
			crashdump_map = &silkworm_maps[1];
			crashdump_part = SATURN_CRASH_DUMP_PARTITION;
			break;
		case SWBD46:	
			printk(KERN_NOTICE "silkworm: "
				"Using SWBD46 flash configuration\n");
			/* Three separate mappings, one NULL terminator */
			if(!(silkworm_maps = kmalloc(
				sizeof(struct my_mtd_mapping_t) * 4, 
					GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, 
				sizeof(struct my_mtd_mapping_t) * 4);
			silkworm_maps[0].map	= &saturn_cfi_boot_map;
			silkworm_maps[0].parts	= &saturn_cfi_boot_parts[0];
			silkworm_maps[0].nparts	= SATURN_JEDEC_NUM_PARTS;
			silkworm_maps[0].probe	= "cfi_probe";
			silkworm_maps[0].base = (void *)ioremap_nocache(
					SATURN_MTD_BOOT_CFI_START,
					SATURN_MTD_BOOT_CFI_LEN);
			cpld_addr = ioremap_nocache(
					SATURN_CPLD_BASE_ADDRESS,
					SATURN_CPLD_SIZE);
			jedec_vpp_offset = 
				SATURN_CPLD_JEDEC_WRITE_ENABLE_OFFSET;
			jedec_vpp_mask = 
				SATURN_CPLD_JEDEC_WRITE_ENABLE_BITS;
			break;
		case SWBD23:	
			pulsar32 = 0;
			/* fall through */
		case SWBD32:	
			printk(KERN_NOTICE "silkworm: "
				"Using %s flash configuration\n", pulsar32 ?
					"SWBD32" : "SWBD23");

			/* Three separate mappings, one NULL terminator */
			if(!(silkworm_maps = kmalloc(sizeof(struct my_mtd_mapping_t) * 4, GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, sizeof(struct my_mtd_mapping_t) * 4);
			silkworm_maps[0].map	= &pulsar_cfi_boot_map;
			silkworm_maps[0].parts	= &pulsar_cfi_boot_parts[0];
			silkworm_maps[0].nparts	= PULSAR_CFI_BOOT_NUM_PARTS;
			silkworm_maps[0].probe	= "cfi_probe";
			silkworm_maps[1].map	= &pulsar_cfi_main_low_map;
			silkworm_maps[1].parts	= &pulsar_cfi_main_low_parts[0];
			silkworm_maps[1].nparts	= PULSAR_CFI_MAIN_LOW_NUM_PARTS;
			silkworm_maps[1].probe	= "cfi_probe";
			silkworm_maps[2].map	= &pulsar_cfi_main_high_map;
			silkworm_maps[2].parts	= &pulsar_cfi_main_high_parts[0];
			silkworm_maps[2].nparts	= PULSAR_CFI_MAIN_HIGH_NUM_PARTS;
			silkworm_maps[2].probe	= "cfi_probe";

			silkworm_maps[0].base	= (void *)ioremap_nocache(	/* 1st cfi device */
					PULSAR_MTD_BOOT_CFI_START,
					PULSAR_MTD_BOOT_CFI_LEN);
			silkworm_maps[1].base = (void *)ioremap_nocache(		/* 2nd cfi device */
					PULSAR_MTD_MAIN_LOW_CFI_START,
					PULSAR_MTD_MAIN_LOW_CFI_LEN);
			silkworm_maps[2].base = (void *)ioremap_nocache(		/* 3rd cfi device */
					PULSAR_MTD_MAIN_HIGH_CFI_START,
					PULSAR_MTD_MAIN_HIGH_CFI_LEN);
			cpld_addr = ioremap_nocache(
					PULSAR_CPLD_BASE_ADDRESS,
					PULSAR_CPLD_SIZE);
			cfi_vpp_offset = PULSAR_CPLD_CFI_WRITE_ENABLE_OFFSET;
			cfi_vpp_mask = PULSAR_CPLD_CFI_WRITE_ENABLE_BITS;
			jedec_vpp_offset = PULSAR_CPLD_BOOT_WRITE_ENABLE_OFFSET;	/* not really jedec */
			jedec_vpp_mask = PULSAR_CPLD_BOOT_WRITE_ENABLE_BITS;
			crashdump_map = &silkworm_maps[1];	/* use pulsar_cfi_low_map */
			crashdump_part = PULSAR_CRASH_DUMP_PARTITION;
			break;
		/* Scimitar-Freescale has only 4MB of boot Flash */
		case SWBD64:	
		case SWBD55:
		case SWBD133:	
			printk(KERN_NOTICE "silkworm: "
				"Using SWBD55 flash configuration\n");

			/* Three separate mappings, one NULL terminator */
			if(!(silkworm_maps = kmalloc(sizeof(struct my_mtd_mapping_t) * 2, GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, sizeof(struct my_mtd_mapping_t) * 2);
			silkworm_maps[0].map	= &scimitar_cfi_boot_map;
			silkworm_maps[0].parts	= &scimitar_cfi_boot_parts[0];
			silkworm_maps[0].nparts	= SCIMITAR_CFI_BOOT_NUM_PARTS;
			silkworm_maps[0].probe	= "cfi_probe";

			silkworm_maps[0].base	= (void *)ioremap_nocache(	/* 1st cfi device */
					SCIMITAR_MTD_BOOT_CFI_START,
					SCIMITAR_MTD_BOOT_CFI_LEN);
			/* CPLD is same as PULSAR */
			cpld_addr = ioremap_nocache(
					PULSAR_CPLD_BASE_ADDRESS,
					PULSAR_CPLD_SIZE);
			cfi_vpp_offset = PULSAR_CPLD_CFI_WRITE_ENABLE_OFFSET;
			cfi_vpp_mask = PULSAR_CPLD_CFI_WRITE_ENABLE_BITS;
			jedec_vpp_offset = PULSAR_CPLD_BOOT_WRITE_ENABLE_OFFSET;	/* not really jedec */
			jedec_vpp_mask = PULSAR_CPLD_BOOT_WRITE_ENABLE_BITS;
			break;

		case SWBD62:	/* Neptune CP (Zentron) */
		case SWBD77:	/* Pluto CP (Zentron) */
		case SWBD63:	/* Neptune CP (Zentron) CO-Processor */
		case SWBD67:	/*MACE */
		case SWBD68:			
		case SWBD79:	/* Spear CP */
		case SWBD89:	/* Striker */
			printk(KERN_NOTICE "silkworm: "
				"Using SWBD62/63/77/67/76/68/79  flash configuration\n");

			/* Three separate mappings, one NULL terminator */
			if(!(silkworm_maps = kmalloc(sizeof(struct my_mtd_mapping_t) * 2, GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, sizeof(struct my_mtd_mapping_t) * 2);
			silkworm_maps[0].map	= &scimitar_cfi_boot_map;
			silkworm_maps[0].parts	= &zentron_cfi_boot_parts[0];
			silkworm_maps[0].nparts	= ZENTRON_CFI_BOOT_NUM_PARTS;
			silkworm_maps[0].probe	= "cfi_probe";

			silkworm_maps[0].base	= (void *)ioremap_nocache(	/* 1st cfi device */
					SCIMITAR_MTD_BOOT_CFI_START,
					SCIMITAR_MTD_BOOT_CFI_LEN);
			/* CPLD is same as PULSAR */
			cpld_addr = ioremap_nocache(
					PULSAR_CPLD_BASE_ADDRESS,
					PULSAR_CPLD_SIZE);
			cfi_vpp_offset = PULSAR_CPLD_CFI_WRITE_ENABLE_OFFSET;
			cfi_vpp_mask = PULSAR_CPLD_CFI_WRITE_ENABLE_BITS;
			jedec_vpp_offset = PULSAR_CPLD_BOOT_WRITE_ENABLE_OFFSET;	/* not really jedec */
			jedec_vpp_mask = PULSAR_CPLD_BOOT_WRITE_ENABLE_BITS;
			break;

		case SWBD76:	/*ELARA */
			printk(KERN_NOTICE "silkworm: "
				"Using SWBD76  flash configuration\n");

			/* Three separate mappings, one NULL terminator */
			if(!(silkworm_maps = kmalloc(sizeof(struct my_mtd_mapping_t) * 2, GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, sizeof(struct my_mtd_mapping_t) * 2);
			silkworm_maps[0].map	= &elara_cfi_boot_map;
			silkworm_maps[0].parts	= &zentron_cfi_boot_parts[0];
			silkworm_maps[0].nparts	= ZENTRON_CFI_BOOT_NUM_PARTS;
			silkworm_maps[0].probe	= "cfi_probe";

			silkworm_maps[0].base	= (void *)ioremap_nocache(	/* 1st cfi device */
					SCIMITAR_MTD_BOOT_CFI_START,
#ifdef CONFIG_MTD_SILKWORM_BOOTSEL_NOMAP
					(SCIMITAR_MTD_BOOT_CFI_LEN - zentron_cfi_boot_parts[5].size)
#else
					SCIMITAR_MTD_BOOT_CFI_LEN
#endif
					);
	
			/* CPLD is same as PULSAR */
			cpld_addr = ioremap_nocache(
					PULSAR_CPLD_BASE_ADDRESS,
					PULSAR_CPLD_SIZE);
			cfi_vpp_offset = PULSAR_CPLD_CFI_WRITE_ENABLE_OFFSET;
			cfi_vpp_mask = PULSAR_CPLD_CFI_WRITE_ENABLE_BITS;
			jedec_vpp_offset = ELARA_CPLD_BOOT_WRITE_ENABLE_OFFSET;	/* not really jedec */
			jedec_vpp_mask = ELARA_CPLD_BOOT_WRITE_ENABLE_BITS;
			break;

		/* Initially Stealth will have kernel flash */
		case SWBD34:
		case SWBD29:
		case SWBD44: /* Viking */
		case SWBD37:
		case SWBD43:
		case SWBD45:
		case SWBD51:
		case SWBD58:
		case SWBD61:
		case SWBD71: /* Tomahawk(440EPX) */
		case SWBD72: /* Bulova(440EPX) */
		case SWBD73: /* Blazer3(Dirk) */
		case SWBD82: /* Utah(440EPX) */
		case SWBD70:
		case SWBD66: /* Spirit(440EPX) */
		case SWBD75: /* Superhawk2 */
		case SWBD78: /* Leda (440EPX) */
		case SWBD88: /* Europa */
		case SWBD83: /* Spike */
		case SWBD85:
		case SWBD86: /* Falcon2 */
		case SWBD87: /* Kestrel2 */
		case SWBD92: /* Gladius(440EPX) */
		case SWBD109: /* Stinger(440EPX) */
		case SWBD117: /* Pharos(440EPX) */
		case SWBD118: /* Tomtoo(440EPX) */
		case SWBD129: /* Bulova2(440EPX) */
		case SWBD130: /* Superhawk3(440EPX) */
		case SWBD124: /* Dragon2E(440EPX) */
		case SWBD125: /* Dragon2T(440EPX) */
		case SWBD134: /* QiLin(440EPX) */
		case SWBD149: /* Kestrel3 */
		case SWBD150: /* Falcon3 */
			printk(KERN_NOTICE "silkworm: "
				"Using SWBD34 flash configuration\n");

			/* One mapping */
			if(!(silkworm_maps = kmalloc(sizeof(struct my_mtd_mapping_t) * 2, GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, sizeof(struct my_mtd_mapping_t) * 2);
			silkworm_maps[0].map	= &pulsar_cfi_boot_map;
			silkworm_maps[0].parts	= &pulsar_cfi_boot_parts[0];
			silkworm_maps[0].nparts	= PULSAR_CFI_BOOT_NUM_PARTS;
			silkworm_maps[0].probe	= "cfi_probe";

			silkworm_maps[0].base	= (void *)ioremap_nocache(	/* 1st cfi device */
					PULSAR_MTD_BOOT_CFI_START,
					PULSAR_MTD_BOOT_CFI_LEN);
			cpld_addr = ioremap_nocache(
					PULSAR_CPLD_BASE_ADDRESS,
					PULSAR_CPLD_SIZE);
			cfi_vpp_offset = PULSAR_CPLD_CFI_WRITE_ENABLE_OFFSET;
			cfi_vpp_mask = PULSAR_CPLD_CFI_WRITE_ENABLE_BITS;
			jedec_vpp_offset = PULSAR_CPLD_BOOT_WRITE_ENABLE_OFFSET;	/* not really jedec */
			jedec_vpp_mask = PULSAR_CPLD_BOOT_WRITE_ENABLE_BITS;
			break;
		case SWBD36:	
		case SWBD47:	
		case SWBD50:
			printk(KERN_NOTICE "silkworm: "
				"Using SWBD36/47/50 flash configuration\n");

			/* One mapping */
			if(!(silkworm_maps = kmalloc(sizeof(struct my_mtd_mapping_t) 
					* 2, GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, sizeof(struct my_mtd_mapping_t) * 2);
			silkworm_maps[0].map	= &marathon_cfi_boot_map;
			silkworm_maps[0].parts	= &marathon_cfi_boot_parts[0];
			silkworm_maps[0].nparts	= MARATHON_CFI_BOOT_NUM_PARTS;
			silkworm_maps[0].probe	= "cfi_probe";

			/* 1st cfi device */
			silkworm_maps[0].base = (void *)ioremap_nocache(
					PULSAR_MTD_BOOT_CFI_START,
					PULSAR_MTD_BOOT_CFI_LEN);
			cpld_addr = ioremap_nocache(
					PULSAR_CPLD_BASE_ADDRESS,
					PULSAR_CPLD_SIZE);

			/* not really jedec */
			jedec_vpp_offset32 = MARATHON_CPLD_BOOT_WRITE_ENABLE_OFFSET;	
			jedec_vpp_mask32 = MARATHON_CPLD_BOOT_WRITE_ENABLE_BITS;
			jedec_vpp_magic32 = MARATHON_CPLD_BOOT_WRITE_MAGIC;
			break;
		case SWBD24:	
			printk(KERN_NOTICE "silkworm: "
				"Using SWBD24 flash configuration\n");

			/* Three separate mappings, one NULL terminator */
			if(!(silkworm_maps = kmalloc(sizeof(struct my_mtd_mapping_t) * 4, GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, sizeof(struct my_mtd_mapping_t) * 4);
			silkworm_maps[0].map	= &athena_cfi_boot_map;
			silkworm_maps[0].parts	= &athena_cfi_boot_parts[0];
			silkworm_maps[0].nparts	= ATHENA_CFI_BOOT_NUM_PARTS;
			silkworm_maps[0].probe	= "cfi_probe";
			silkworm_maps[1].map	= &athena_cfi_main_low_map;
			silkworm_maps[1].parts	= &athena_cfi_main_low_parts[0];
			silkworm_maps[1].nparts	= ATHENA_CFI_MAIN_LOW_NUM_PARTS;
			silkworm_maps[1].probe	= "cfi_probe";
			silkworm_maps[2].map	= &athena_cfi_main_high_map;
			silkworm_maps[2].parts	= &athena_cfi_main_high_parts[0];
			silkworm_maps[2].nparts	= ATHENA_CFI_MAIN_HIGH_NUM_PARTS;
			silkworm_maps[2].probe	= "cfi_probe";

			silkworm_maps[0].base	= (void *)ioremap_nocache(
/* 1st cfi device */
					ATHENA_MTD_BOOT_CFI_START,
					ATHENA_MTD_BOOT_CFI_LEN);
			silkworm_maps[1].base = (void *)ioremap(			/* 2nd cfi device */
					ATHENA_MTD_MAIN_LOW_CFI_START,
					ATHENA_MTD_MAIN_LOW_CFI_LEN);
			silkworm_maps[2].base = (void *)ioremap(			/* 3rd cfi device */
					ATHENA_MTD_MAIN_HIGH_CFI_START,
					ATHENA_MTD_MAIN_HIGH_CFI_LEN);
			cpld_addr = ioremap_nocache(
					ATHENA_CPLD_BASE_ADDRESS,
					ATHENA_CPLD_SIZE);
			cfi_vpp_offset = ATHENA_CPLD_CFI_WRITE_ENABLE_OFFSET;
			cfi_vpp_mask = ATHENA_CPLD_CFI_WRITE_ENABLE_BITS;
			jedec_vpp_offset = ATHENA_CPLD_BOOT_WRITE_ENABLE_OFFSET;	/* not really jedec */
			jedec_vpp_mask = ATHENA_CPLD_BOOT_WRITE_ENABLE_BITS;
			crashdump_map = &silkworm_maps[1];	/* use athena_cfi_low_map */
			crashdump_part = ATHENA_CRASH_DUMP_PARTITION;
			break;
		case SWBD39:	
			printk(KERN_NOTICE "silkworm: Using swbd39 (Forte)"
				" flash configuration\n");

			if(!(silkworm_maps = kmalloc(sizeof(struct my_mtd_mapping_t) * 3, GFP_KERNEL))) {
				printk("Failed to allocate my_mtd_mapping\n");
				return -EIO;
			}
			memset(silkworm_maps, 0, sizeof(struct my_mtd_mapping_t) * 3);

			silkworm_maps[0].map    = &forte_cfi_low_map;
			silkworm_maps[0].parts  = &forte_cfi_low_parts[0];
			silkworm_maps[0].nparts = FORTE_CFI_LOW_NUM_PARTS;
			silkworm_maps[0].probe  = "cfi_probe";

			silkworm_maps[1].map    = &forte_cfi_high_map;
			silkworm_maps[1].parts  = &forte_cfi_high_parts[0];
			silkworm_maps[1].nparts = FORTE_CFI_HIGH_NUM_PARTS;
			silkworm_maps[1].probe  = "cfi_probe";

			silkworm_maps[0].base = (void *)ioremap_nocache(
				FORTE_MTD_CFI_LOW_START,
				FORTE_MTD_CFI_LOW_LEN);

			silkworm_maps[1].base = (void *)ioremap_nocache(
				FORTE_MTD_CFI_HIGH_START,
				FORTE_MTD_CFI_HIGH_LEN);

			cpld_addr = ioremap_nocache(
					FORTE_CPLD_BASE_ADDRESS,
					FORTE_CPLD_SIZE);

			cfi_vpp_offset32  = FORTE_CPLD_CFI_WRITE_ENABLE_OFFSET;
			cfi_vpp_mask32    = FORTE_CPLD_CFI_WRITE_ENABLE_BITS;
			jedec_vpp_offset  = 0;
			jedec_vpp_mask    = 0;

			crashdump_map  = &silkworm_maps[1];
			crashdump_part = FORTE_CRASH_DUMP_PARTITION;

			break;
	}

	/* Add all the mapped partitions */
	idx = 0;
	while(silkworm_maps[idx].map != NULL) {

		int nparts = silkworm_maps[idx].nparts;
		void *base = silkworm_maps[idx].base;
		struct map_info *map = silkworm_maps[idx].map;
		struct mtd_partition *parts = silkworm_maps[idx].parts;
		struct mtd_info *my_mtd;
		char *probe_name = silkworm_maps[idx].probe;

		if(((void __iomem *)map->map_priv_1 = base) == NULL) {
			printk("Failed to ioremap mtd base address\n");
			return -EIO;
		}
		silkworm_maps[idx].mtd_handle = my_mtd = do_map_probe(probe_name, map);
		if (my_mtd) {
			// my_mtd->module = THIS_MODULE;
			add_mtd_partitions(my_mtd, parts, nparts);
		}
		else {
			printk(KERN_NOTICE "Flash device [%s] NOT initialized!\n", map->name);
			iounmap((void *)map->map_priv_1);
			ret = -ENXIO;
		}
		idx++;
	}

	/*
	 * Even though flash dumper is disabled, this flash_dump_init() call
	 * is required for refresh_watchdog()
	 */
	if (!crashdump_map) {
		flash_dump_init(0, 0, 2, 0, NULL, 0, 0, 0);
	}

	/* Today we only use one of the crash dump partitions */
	/* Initialize the flash panic-dumper code */
	if(ret == 0 && crashdump_map) {
		unsigned long crashdump_addr = crashdump_map->base + crashdump_map->parts[crashdump_part].offset;
		unsigned long crashdump_size = crashdump_map->parts[crashdump_part].size;
		unsigned long erasesize = crashdump_map->mtd_handle->erasesize;
		char *crashdump_name = crashdump_map->parts[crashdump_part].name;
		int crashdump_buswidth = crashdump_map->map->bankwidth;

		/* Assumption:  If multiple CFI flash devices exist, and flashdumps are done
		 * on multiple CFI devices, all CFI devices must have the same erase size.
		 */
		flash_dump_init(crashdump_addr, crashdump_size, crashdump_buswidth,
				erasesize, crashdump_name, cpld_addr, cfi_vpp_offset, cfi_vpp_mask);
	} 

	return ret;
}

static void __exit cleanup_silkworm(void)
{
	int idx = 0;

	/* destroy all the mapped partitions */
	while(silkworm_maps != NULL && silkworm_maps[idx].map != NULL) {

		if(silkworm_maps[idx].mtd_handle) {
			del_mtd_partitions(silkworm_maps[idx].mtd_handle);
			map_destroy(silkworm_maps[idx].mtd_handle);
		}
		if(silkworm_maps[idx].base) {
			iounmap((void __iomem *)silkworm_maps[idx].map->map_priv_1);
			(void __iomem *)silkworm_maps[idx].map->map_priv_1 = NULL;
		}
		idx++;
	}
	if(cpld_addr)
		iounmap((void *)cpld_addr);

}

module_init(init_silkworm);
module_exit(cleanup_silkworm);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("B. Kuschak, Brocade");
MODULE_DESCRIPTION("MTD map driver for Silkworm family");
